########################################################################
# File name: utils.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import functools

import PyQt5.Qt as Qt


if not hasattr(asyncio, "ensure_future"):
    asyncio.ensure_future = getattr(asyncio, "async")


def asyncified_done(task):
    task.result()


def asyncified_unblock(dlg, cursor, task):
    dlg.setCursor(cursor)
    dlg.setEnabled(True)


def asyncify(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        task = asyncio.ensure_future(fn(*args, **kwargs))
        task.add_done_callback(asyncified_done)
    return wrapper


def asyncify_blocking(*, cursor=Qt.Qt.WaitCursor):
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            prev_cursor = self.cursor()
            self.setEnabled(False)
            self.setCursor(cursor)
            try:
                task = asyncio.async(fn(self, *args, **kwargs))
            except:
                self.setEnabled(True)
                self.setCursor(prev_cursor)
                raise
            task.add_done_callback(asyncified_done)
            task.add_done_callback(functools.partial(
                asyncified_unblock,
                self, prev_cursor))

        return wrapper

    return decorator


@asyncio.coroutine
def exec_async(dlg, set_modal=Qt.Qt.WindowModal):
    future = asyncio.Future()

    def done(result):
        nonlocal future
        future.set_result(result)

    dlg.finished.connect(done)
    if set_modal is not None:
        dlg.windowModality = set_modal
    dlg.show()
    try:
        return (yield from future)
    except asyncio.CancelledError:
        print("being cancelled, rejecting dialogue and re-raising")
        dlg.finished.disconnect(done)
        dlg.reject()
        raise



########################################################################
# File name: __init__.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import argparse
import logging
import signal
import sys

import quamash

import PyQt5.Qt as Qt

import cuteborg.backend
import cuteborg.config

from .ui_main import Ui_MainWindow


class RepositoriesModel(Qt.QAbstractTableModel):
    COLUMN_ID = 0
    COLUMN_NAME = 1
    COLUMN_TYPE = 2
    COLUMN_ENCRYPTION = 3
    COLUMN_PATH = 4
    COLUMN_DEVICE = 5
    COLUMN_COUNT = 6

    def __init__(self):
        super().__init__()
        self.repositories = []

    def set_repositories(self, repositories):
        self.beginResetModel()
        self.repositories[:] = repositories.values()
        self.endResetModel()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        return len(self.repositories)

    def columnCount(self, parent):
        return self.COLUMN_COUNT

    def headerData(self, section, orientation, role):
        if orientation != Qt.Qt.Horizontal:
            return
        if role == Qt.Qt.DisplayRole:
            return [
                "ID",
                "Name",
                "Type",
                "Encryption",
                "Path",
                "Device / Host",
            ][section]

    def index(self, row, column, parent):
        if parent.isValid():
            return Qt.QModelIndex()

        return self.createIndex(
            row,
            column,
            self.repositories[row],
        )

    def data(self, index, role):
        if not index.isValid():
            return

        if role != Qt.Qt.DisplayRole:
            return

        repo = index.internalPointer()
        col = index.column()

        if col == self.COLUMN_ID:
            return repo.id_
        elif col == self.COLUMN_NAME:
            return repo.name
        elif col == self.COLUMN_TYPE:
            if isinstance(repo, cuteborg.config.LocalRepositoryConfig):
                return "local"
            else:
                return "remote"
        elif col == self.COLUMN_ENCRYPTION:
            return {
                cuteborg.backend.EncryptionMode.NONE: "no encryption",
                cuteborg.backend.EncryptionMode.KEYFILE: "local keyfile",
                cuteborg.backend.EncryptionMode.REPOKEY: "repository keyfile",
            }[repo.encryption_mode]
        elif col == self.COLUMN_PATH:
            return repo.path
        elif col == self.COLUMN_DEVICE:
            if isinstance(repo, cuteborg.config.LocalRepositoryConfig):
                if repo.removable_device_uuid is not None:
                    return "/dev/disk/by-uuid/"+repo.removable_device_uuid
                else:
                    return ""
            else:
                parts = ["ssh://"]
                if repo.user:
                    parts.append(repo.user)
                    parts.append("@")
                parts.append(repo.host)
                return "".join(parts)


class JobsModel(Qt.QAbstractTableModel):
    COLUMN_NAME = 0
    COLUMN_SCHEDULE = 1
    COLUMN_COMPRESSION = 2
    COLUMN_COUNT = 3

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.jobs = []

    def set_jobs(self, jobs):
        self.beginResetModel()
        self.jobs[:] = jobs.values()
        self.endResetModel()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        return len(self.jobs)

    def columnCount(self, parent):
        return self.COLUMN_COUNT

    def headerData(self, section, orientation, role):
        if orientation != Qt.Qt.Horizontal:
            return
        if role == Qt.Qt.DisplayRole:
            return [
                "Name",
                "Schedule",
                "Compression",
            ][section]

    def index(self, row, column, parent):
        if parent.isValid():
            return Qt.QModelIndex()

        return self.createIndex(
            row,
            column,
            self.jobs[row],
        )

    def data(self, index, role):
        if not index.isValid():
            return

        if role != Qt.Qt.DisplayRole:
            return

        job = index.internalPointer()
        col = index.column()

        if col == self.COLUMN_NAME:
            return job.name
        elif col == self.COLUMN_SCHEDULE:
            schedule = job.schedule or self.config.schedule
            return "every {} {}(s)".format(
                schedule.interval_step,
                schedule.interval_unit.value,
            )
        elif col == self.COLUMN_COMPRESSION:
            if job.compression_method is None:
                return "uncompressed"
            else:
                return job.compression_method.value


class MainWindow(Qt.QMainWindow):
    def __init__(self, loop, logger):
        super().__init__()
        self.loop = loop
        self.logger = logger

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.config = cuteborg.config.Config.from_raw(
            cuteborg.config.load_config()
        )

        self.repositories_model = RepositoriesModel()
        self.repositories_model.set_repositories(
            self.config.repositories
        )
        proxy_model = Qt.QSortFilterProxyModel(self.ui.repositories_view)
        proxy_model.setSourceModel(self.repositories_model)
        proxy_model.setSortCaseSensitivity(Qt.Qt.CaseInsensitive)
        proxy_model.setSortLocaleAware(True)
        proxy_model.setSortRole(Qt.Qt.DisplayRole)
        proxy_model.sort(
            RepositoriesModel.COLUMN_NAME,
            Qt.Qt.AscendingOrder,
        )
        self.ui.repositories_view.setModel(
            proxy_model,
        )
        self.ui.repositories_view.horizontalHeader().setSectionResizeMode(
            Qt.QHeaderView.ResizeToContents
        )
        self.ui.repositories_view.horizontalHeader().setSortIndicator(
            RepositoriesModel.COLUMN_NAME,
            Qt.Qt.AscendingOrder,
        )

        self.jobs_model = JobsModel(self.config)
        self.jobs_model.set_jobs(
            self.config.jobs
        )
        proxy_model = Qt.QSortFilterProxyModel(self.ui.jobs_view)
        proxy_model.setSourceModel(self.jobs_model)
        proxy_model.setSortCaseSensitivity(Qt.Qt.CaseInsensitive)
        proxy_model.setSortLocaleAware(True)
        proxy_model.setSortRole(Qt.Qt.DisplayRole)
        proxy_model.sort(
            JobsModel.COLUMN_NAME,
            Qt.Qt.AscendingOrder,
        )
        self.ui.jobs_view.setModel(
            proxy_model,
        )
        self.ui.jobs_view.horizontalHeader().setSectionResizeMode(
            Qt.QHeaderView.ResizeToContents
        )
        self.ui.jobs_view.horizontalHeader().setSortIndicator(
            JobsModel.COLUMN_NAME,
            Qt.Qt.AscendingOrder,
        )


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Increase verbosity (up to -vvv)"
    )

    args = parser.parse_args()

    logging.basicConfig(
        level={
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
        }.get(args.verbosity, logging.DEBUG),
    )

    logger = logging.getLogger("qtborg.manager")

    logging.getLogger("quamash").setLevel(logging.INFO)

    Qt.QApplication.setAttribute(Qt.Qt.AA_UseHighDpiPixmaps)
    app = Qt.QApplication(sys.argv[:1])
    asyncio.set_event_loop(quamash.QEventLoop(app=app))

    signal.signal(
        signal.SIGINT,
        signal.SIG_DFL,
    )

    signal.signal(
        signal.SIGTERM,
        signal.SIG_DFL,
    )

    loop = asyncio.get_event_loop()
    try:
        main = MainWindow(loop, logger)
        main.show()
        rc = loop.run_forever()
        del main
        del app
        import gc
        gc.collect()
        sys.exit(rc)
    finally:
        loop.close()



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_edit_job.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(629, 451)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(Dialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.tabWidget.addTab(self.tab_2, "")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.tabWidget.addTab(self.tab_3, "")
        self.tab_4 = QtWidgets.QWidget()
        self.tab_4.setObjectName("tab_4")
        self.tabWidget.addTab(self.tab_4, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(0)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("Dialog", "Repositories"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("Dialog", "Sources"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), _translate("Dialog", "Schedule"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), _translate("Dialog", "Pruning"))




########################################################################
# File name: cli.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import asyncio
import logging
import pathlib
import sys

import xdg.BaseDirectory

import PyQt5.Qt as Qt

import quamash

from . import Main


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Increase verbosity (up to -vvv)"
    )

    parser.add_argument(
        "--socket-path",
        default=None,
        type=pathlib.Path,
        help="Override the socket path "
        "(defaults to a user-specific socket)"
    )

    args = parser.parse_args()

    logging.basicConfig(
        level={
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
        }.get(args.verbosity, logging.DEBUG),
    )

    if args.socket_path is None:
        args.socket_path = pathlib.Path(
            xdg.BaseDirectory.get_runtime_dir(strict=False)
        ) / "cuteborg" / "scheduler.sock"

    logger = logging.getLogger("qtborg.trayicon")

    logging.getLogger("quamash").setLevel(logging.INFO)

    Qt.QApplication.setAttribute(Qt.Qt.AA_UseHighDpiPixmaps)
    app = Qt.QApplication(sys.argv[:1])
    app.setQuitOnLastWindowClosed(False)
    asyncio.set_event_loop(quamash.QEventLoop(app=app))

    loop = asyncio.get_event_loop()
    try:
        iconapp = Main(loop, logger, args.socket_path)
        rc = loop.run_until_complete(iconapp.run())
        del iconapp
        del app
        import gc
        gc.collect()
        sys.exit(rc)
    finally:
        loop.close()



########################################################################
# File name: status.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import logging

from . import protocol


@asyncio.coroutine
def get_status(loop, socket_path):
    logger = logging.getLogger("cuteborg.scheduler")

    try:
        endpoint = yield from asyncio.wait_for(
            protocol.test_and_get_socket(
                loop,
                logger,
                socket_path,
            ),
            timeout=5
        )
    except (asyncio.TimeoutError, FileNotFoundError, ConnectionError) as exc:
        msg = str(exc)
        if isinstance(exc, asyncio.TimeoutError):
            msg = "timeout"
        logger.error("failed to connect to scheduler at %r: %s",
                     str(socket_path),
                     msg)
        raise RuntimeError(
            "failed to connect to scheduler"
        )

    response, data = yield from endpoint.request(
        protocol.ToplevelCommand.STATUS
    )

    return data



########################################################################
# File name: protocol.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import enum
import struct
import toml


ExtHeader = struct.Struct(
    # version
    # size (16 bits)
    "<BH",
)


class ProtocolState(enum.Enum):
    TOPLEVEL = 0
    RECV_EXT_HEADER = 1
    RECV_EXT_BODY = 2


class ToplevelCommand(enum.Enum):
    PING = b"AYT?"  # Are You There?
    PONG = b"YSH!"  # Yes, i’m Still Here!

    RESCHEDULE = b"RLS!"  # ReLoad and reSchedule!

    POLL = b"KMP!"  # Keep Me Posted!

    STATUS = b"WIU?"  # What Is Up?

    ERROR = b"ERR!"  # ERRor

    OKAY = b"OKAY"  # OKAY!

    EXTENDED_REQUEST = b"EXT:"  # EXTended

    EXIT = b"EXIT"  # EXIT


class ControlProtocol(asyncio.Protocol):
    def __init__(self, logger, handler):
        super().__init__()

        self.logger = logger
        self.handler = handler

        self.breakdown_future = asyncio.Future()

        self._buffer = []
        self._buffer_len = 0

        self._ext_size = 0
        self._ext_cmd = None

        self._state = ProtocolState.TOPLEVEL

    def _squash_buffer(self, extract_bytes=0):
        buf = b"".join(self._buffer)
        assert len(buf) == self._buffer_len
        assert len(buf) >= extract_bytes
        result = buf[:extract_bytes]
        self._buffer = [buf[extract_bytes:]]
        self._buffer_len -= extract_bytes
        return result

    def _process(self, state):
        if state == ProtocolState.TOPLEVEL:
            if self._buffer_len < 5:
                return state

            cmd = self._squash_buffer(5)

            self.logger.debug("received command bytes: %r", cmd)

            if cmd[4] != b"\n"[0]:
                raise ValueError(
                    "commands must be terminated with a newline"
                )

            actual_cmd = ToplevelCommand(cmd[:4])

            self.logger.debug("received command: %r", actual_cmd)

            if actual_cmd.value.endswith(b":"):
                self.logger.debug("command is extended, "
                                  "waiting for more data")
                self._ext_cmd = actual_cmd
                return ProtocolState.RECV_EXT_HEADER

            self._process_command(actual_cmd)
            return state

        elif state == ProtocolState.RECV_EXT_HEADER:
            if self._buffer_len < ExtHeader.size:
                return state

            header = self._squash_buffer(ExtHeader.size)
            version, size = ExtHeader.unpack(header)

            if version != 1:
                raise ValueError("unknown version: {!r}".format(version))

            self._ext_size = size

            return ProtocolState.RECV_EXT_BODY

        elif state == ProtocolState.RECV_EXT_BODY:
            if self._buffer_len < self._ext_size:
                return state

            raw_data = self._squash_buffer(self._ext_size).decode("utf-8")
            try:
                data = toml.loads(raw_data)
            except:
                raise ValueError("failed to decode message: {!r}".format(
                    raw_data
                ))

            self._process_command(
                self._ext_cmd,
                data,
            )

            return ProtocolState.TOPLEVEL

    def _error(self):
        if hasattr(self, "_transport"):
            self.send(
                ToplevelCommand.ERROR,
            )
            self._transport.close()

    def _process_command(self, cmd, extra_data=None):
        self.logger.debug("%r (extra_data=%r)", cmd, extra_data)
        if cmd == ToplevelCommand.PING:
            self._transport.writelines([
                ToplevelCommand.PONG.value,
                b"\n"
            ])
        else:
            result = self.handler(self, cmd, extra_data)
            if result is not None:
                toplevel, extra_data = result
                self.send(toplevel, extra_data)

    def send(self, cmd, extra_data=None):
        if not hasattr(self, "_transport"):
            raise ConnectionError("disconnected")

        self.logger.debug("sending %r (extra_data=%r)", cmd, extra_data)

        parts = [
            cmd.value,
            b"\n",
        ]

        if extra_data is not None:
            if not cmd.value.endswith(b":"):
                raise ValueError(
                    "that command does not support extended data"
                )
            serialised = toml.dumps(extra_data).encode("utf-8")
            parts.extend([
                ExtHeader.pack(1, len(serialised)),
                serialised,
            ])

        elif cmd.value.endswith(b":") and extra_data is None:
            raise ValueError(
                "that command requires extended data"
            )

        self._transport.writelines(parts)

    def connection_made(self, transport):
        self._transport = transport

    def connection_lost(self, exc):
        if exc is not None:
            self.breakdown_future.set_exception(exc)
        else:
            self.breakdown_future.set_result(None)
        del self._transport

    def data_received(self, data):
        self.logger.debug("recv: %r", data)

        self._buffer.append(data)
        self._buffer_len += len(data)

        try:
            new_state = self._process(self._state)
            try:
                while new_state != self._state:
                    self._state = new_state
                    new_state = self._process(new_state)
            finally:
                self._state = new_state
        except:
            self.logger.error(
                "while processing input",
                exc_info=True,
            )
            self._error()

    def eof_received(self):
        pass

    def close(self):
        self._transport.close()


class ClientProtocolEndpoint:
    protocol = None

    def __init__(self):
        super().__init__()
        self.response_futs = []
        self.poll_future = None

    def close(self):
        if self.protocol is not None:
            self.protocol.close()

    @asyncio.coroutine
    def request(self, cmd, extra_data=None, timeout=None):
        if self.response_futs:
            raise RuntimeError("there is a command pending")
        if self.poll_future is not None:
            raise RuntimeError("there is a poll pending")

        fut = asyncio.Future()
        self.response_futs.append(fut)
        self.protocol.send(cmd, extra_data)

        done, pending = yield from asyncio.wait(
            [
                fut,
                self.protocol.breakdown_future,
            ],
            return_when=asyncio.FIRST_COMPLETED,
            timeout=timeout
        )

        if fut in done:
            return fut.result()

        self.protocol.breakdown_future.result()
        raise ConnectionError("unknown connection error")

    @asyncio.coroutine
    def poll_start(self):
        if self.response_futs:
            raise RuntimeError("there is a command pending")
        if self.poll_future is not None:
            raise RuntimeError("there is a poll pending")

        fut1 = asyncio.Future()
        self.poll_future = asyncio.Future()

        self.response_futs.append(fut1)
        self.response_futs.append(self.poll_future)

        self.protocol.send(ToplevelCommand.POLL)

        done, pending = yield from asyncio.wait(
            [
                fut1,
                self.protocol.breakdown_future,
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if fut1 in done:
            return fut1.result()

        self.protocol.breakdown_future.result()
        raise ConnectionError("unknown connection error")

    @asyncio.coroutine
    def poll_finalise(self):
        if self.poll_future is None:
            raise RuntimeError("there is no poll pending")

        done, pending = yield from asyncio.wait(
            [
                self.poll_future,
                self.protocol.breakdown_future,
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if self.poll_future in done:
            result = self.poll_future.result()
            self.poll_future = None
            return result

        self.poll_future = None

        self.protocol.breakdown_future.result()
        raise ConnectionError("unknown connection error")

    def handle_request(self, protocol, cmd, extra_data):
        assert self.protocol == protocol
        if not self.response_futs:
            return protocol.ToplevelCommand.ERROR, None

        fut = self.response_futs.pop(0)
        fut.set_result((cmd, extra_data))


@asyncio.coroutine
def test_and_get_socket(loop, logger, socket_path):
    endpoint = ClientProtocolEndpoint()

    def protocol_factory():
        nonlocal endpoint

        proto = ControlProtocol(
            logger,
            endpoint.handle_request,
        )
        endpoint.protocol = proto

        return proto

    try:
        _, protocol = yield from loop.create_unix_connection(
            protocol_factory,
            str(socket_path))
    except ConnectionRefusedError:
        # we can unlink the socket
        socket_path.unlink()
        raise FileNotFoundError()

    cmd, _ = yield from endpoint.request(ToplevelCommand.PING)
    if cmd != ToplevelCommand.PONG:
        raise ConnectionError("unexpected response to PING command")

    return endpoint



########################################################################
# File name: verify.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import enum
import subprocess


class ContentDiff(enum.Enum):
    EQUAL = 0
    SIZE_CHANGED = 1
    CHECKSUM_CHANGED = 2


class MetadataDiff(enum.Enum):
    TIMES = 0
    UNIX_PERMISSIONS = 1
    ACLS = 2
    XATTRS = 3
    TYPE = 4


class _RsyncFileFlag(enum.Enum):
    FILE_MISSING_LOCALLY = ">"
    FILE_MISSING_IN_ARCHIVE = "<"
    IS_HARDLINK = "h"
    CONTENT_UNCHANGED = "."
    _MESSAGE = "*"


@asyncio.coroutine
def diff_rsync(a, b, *, checksum=True):
    """
    Calculate the difference between two filesystem subtrees.

    :param a: rsync source path
    :param b: rsync target path

    This is calculated using rsync in dry-run mode with ``--itemize-changes``
    to obtain the information.

    If `checksum` is true, the files are compared using their checksums. This
    is considerably slower, but protects against silent data corruption. If you
    do not trust borg, use `checksum`.
    """

    args = [
        "rsync",
        "-raHAPSEX",  # my preferred recipe
        "--dry-run",  # we don’t want to actually do anything to the files
        "--itemize-changes",  # but we want to know what *would* be done
        "--delete",  # ensure that missing files are reported
    ]

    if checksum:
        args.append("--checksum")

    proc = asyncio.create_subprocess_exec(
        *args,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
    )

    try:
        finished = False
        first = True
        while True:
            try:
                line = yield from proc.stdout.readline()
            except asyncio.streams.IncompleteReadError as exc:
                line = exc.partial
                finished = True

            if first:
                continue



            if finished:
                break
    except asyncio.CancelledError:
        pass

    if proc.returncode is None:
        try:
            proc.terminate()
        except ProcessLookupError:
            pass

    yield from proc.wait()



########################################################################
# File name: jobs.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import asyncio

import pytz

from datetime import datetime

import cuteborg.backend

from . import utils


def _schedule_next(now, schedule, last_run):
    start_of_current_interval = utils.align_to_interval(
        now,
        schedule.interval_step,
        schedule.interval_unit,
    )

    if (last_run is not None and
            last_run >= start_of_current_interval):
        start_of_current_interval = utils.step_interval(
            start_of_current_interval,
            schedule.interval_step,
            schedule.interval_unit,
        )

    # end_of_current_interval = utils.step_interval(
    #     start_of_current_interval,
    #     schedule.interval_step,
    #     schedule.interval_unit,
    # )

    if now > start_of_current_interval:
        return now

    # TODO: honour scheduling preferences
    return start_of_current_interval


class Job(metaclass=abc.ABCMeta):
    def __init__(self, logger, scheduler):
        super().__init__()
        self.logger = logger
        self.scheduler = scheduler

    @abc.abstractproperty
    def key(self):
        pass

    @asyncio.coroutine
    def retry_locked(self, action, *args, **kwargs):
        while True:
            try:
                yield from action(*args, **kwargs)
            except cuteborg.backend.RepositoryLocked:
                self.logger.warning(
                    "repository is locked, will retry after 30 seconds"
                )
                self.scheduler.set_job_error(
                    self.key,
                    "repository is locked by an external process",
                )
                yield from asyncio.sleep(30)
                continue
            except cuteborg.backend.RepositoryUnreachable:
                self.logger.warning(
                    "repository is unreachable, will retry after 30 seconds"
                )
                self.scheduler.set_job_error(
                    self.key,
                    "repository is unreachable",
                )
                yield from asyncio.sleep(30)
                continue
            break

    @abc.abstractmethod
    def exec(self):
        """
        Execute a single instance of the job.

        This coroutine is cancelled when the scheduler shuts down, but not when
        re-scheduling takes place.
        """

    @abc.abstractmethod
    def run(self):
        """
        Schedule the job and run it as needed.

        This coroutine is cancelled when rescheduling takes place. Running jobs
        are protected if ran through :meth:`.core.Scheduler.execute_job`.
        """


class PeriodicRepositoryJob(Job):
    def __init__(self, logger, scheduler, repository_cfg, schedule):
        super().__init__(logger, scheduler)
        self.repository_cfg = repository_cfg
        self.schedule = schedule

    @asyncio.coroutine
    def exec(self, **kwargs):
        # waits until the storage of the repository is reachable
        repo_path = yield from self.scheduler.wait_for_repository_storage(
            self,
            self.repository_cfg
        )

        self.logger.debug(
            "repository is at %r",
            repo_path,
        )

        self.logger.debug("locking repository %r", self.repository_cfg.id_)
        with (yield from self.scheduler.lock_repository(
                self,
                self.repository_cfg.id_)):
            self.logger.debug("locked repository %r", self.repository_cfg.id_)
            yield from self._exec(repo_path=repo_path, **kwargs)

    @asyncio.coroutine
    def run(self):
        while True:
            last_run = self.scheduler.get_last_run(self.key)
            if last_run is None:
                self.logger.debug("job was not run yet")
            else:
                self.logger.debug("last run was at %r", last_run)

            next_run = _schedule_next(
                datetime.utcnow(),
                self.schedule,
                last_run,
            )

            self.logger.debug("next run is at %s", next_run)
            yield from self.scheduler.sleep_until(self, next_run)
            yield from self.scheduler.execute_job(self)


class PruneRepository(PeriodicRepositoryJob):
    def __init__(self, logger, scheduler, prune_cfg, repository_cfg, schedule):
        super().__init__(logger, scheduler, repository_cfg, schedule)
        self.prune_cfg = prune_cfg

    @property
    def key(self):
        return ("prune",
                self.repository_cfg.id_)

    @asyncio.coroutine
    def _exec(self, repo_path):
        job_names = []
        for name, job in self.scheduler.config.jobs.items():
            if self.repository_cfg in job.repositories:
                job_names.append(name)

        settings = self.repository_cfg.prune.to_kwargs()

        context = cuteborg.backend.Context()
        self.repository_cfg.setup_context(context)

        for name in job_names:
            prefix = name + "-"

            self.logger.debug(
                "pruning archives starting with %r", prefix
            )

            self.scheduler.set_job_progress(
                self,
                {
                    "job": name,
                }
            )

            yield from self.retry_locked(
                self.scheduler.backend.prune_repository,
                repo_path,
                context,
                prefix,
                **settings,
            )
            self.scheduler.set_job_error(self.key, None)

        self.scheduler.set_job_progress(self, None)


class CreateArchive(PeriodicRepositoryJob):
    def __init__(self, logger, scheduler, job_cfg, repository_cfg, schedule):
        super().__init__(logger, scheduler, repository_cfg, schedule)
        self.job_cfg = job_cfg

    @property
    def key(self):
        return ("create_archive",
                self.job_cfg.name,
                self.repository_cfg.id_)

    def _update_progress(self, progress):
        self.scheduler.set_job_error(self.key, None)
        if progress is not None:
            self.scheduler.set_job_progress(
                self,
                {
                    key: (" ".join(value)
                          if not isinstance(value, (int, float))
                          else value)
                    for key, value in progress.items()
                }
            )
        else:
            self.scheduler.set_job_progress(self, None)

    @asyncio.coroutine
    def _exec(self, repo_path):
        self.logger.debug("starting job")
        now = datetime.utcnow()

        archive_name = self.job_cfg.name + "-" + now.isoformat()

        context = cuteborg.backend.Context()
        context.progress_callback = self._update_progress
        self.repository_cfg.setup_context(context)
        self.job_cfg.setup_context(context)

        yield from self.retry_locked(
            self.scheduler.backend.create_archive,
            repo_path,
            archive_name,
            self.job_cfg.sources,
            context,
        )



########################################################################
# File name: cli.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import logging
import pathlib
import pprint
import uuid
import sys

from . import config, utils


def cmd_repository_add_local_directory(logger, cfg, state, args):
    for repo in cfg.setdefault("repository", []):
        if repo["type"] != "local":
            continue
        if repo["path"] == str(args.path):
            logger.error(
                "a repository with path %r already exists as id %r",
                str(args.path),
                repo["id"],
            )
            return False, False, 1

    uid = str(uuid.uuid4())

    repository = {
        "id": uid,
        "type": "local_directory",
        "path": str(args.path),
    }

    cfg["repository"].append(repository)

    print(uid)

    return True, False, 0


def cmd_repository_remove(logger, cfg, state, args):
    # TODO: check that no job uses that repository
    ifound = config.find_repository_by_id(
        cfg.get("repository", []),
        args.id,
    )

    if ifound is None:
        logger.error(
            "no such repository: %r",
            args.id,
        )
        return False, False, 1

    del cfg["repository"][ifound]

    return True, False, 0


def cmd_repository_list(logger, cfg, state, args):
    headers = [
        "id",
        "type",
        "path",
        "extra",
    ]

    rows = [
        (
            repo["id"],
            repo["type"],

            repo["local"]["path"] if repo["type"] == "local" else
            repo["remote"]["path"] if repo["type"] == "remote" else "",

            "dev={!r}".format(
                repo["local"].get("removable", {}).get("device_uuid", None),
            ) if repo["type"] == "local" else
            "user={!r} host={!r}".format(
                    repo["remote"].get("user"),
                    repo["remote"].get("host"),
            ) if repo["type"] == "remote" else ""
        )
        for repo in cfg.get("repository", [])
    ]

    utils.print_table(headers, rows)

    return False, False, 0


def cmd_job_add(logger, cfg, state, args):
    pass


def path(s):
    p = pathlib.Path(s).absolute()
    return p


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Increase verbosity (up to -vvv)",
    )

    subparsers = parser.add_subparsers()

    subparser = subparsers.add_parser(
        "repository-add-local-directory",
        help="Add a new directory repository",
    )
    subparser.set_defaults(func=cmd_repository_add_local_directory)
    subparser.add_argument(
        "path",
        type=path,
        help="Path to the local directory to use as repository"
    )

    subparser = subparsers.add_parser(
        "repository-remove",
        help="Remove a repository",
    )
    subparser.set_defaults(func=cmd_repository_remove)
    subparser.add_argument(
        "id",
        help="ID of the repository to remove"
    )

    subparser = subparsers.add_parser(
        "repository-list",
        help="List known repositories",
    )
    subparser.set_defaults(func=cmd_repository_list)

    subparser = subparsers.add_parser(
        "job-add-local",
        help="Add local sourced job"
    )
    subparser.add_argument(
        "--repository",
        action="append",
        help="Add repository to run the job into"
    )

    args = parser.parse_args()

    logging.basicConfig(
        level={
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
        }.get(args.verbosity, logging.DEBUG),
    )

    logger = logging.getLogger("cuteborg")

    cfg = config.load_config()
    state = config.load_state()

    assert hasattr(args, "func")

    cfg_changed, state_changed, rc = args.func(
        logger,
        cfg,
        state,
        args,
    )

    if cfg_changed:
        config.save_config(cfg)
    if state_changed:
        config.save_state(state)

    sys.exit(rc)



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_new_job_wizard.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_NewJobWizard(object):
    def setupUi(self, NewJobWizard):
        NewJobWizard.setObjectName("NewJobWizard")
        NewJobWizard.resize(563, 436)
        self.page_repositories = QtWidgets.QWizardPage()
        self.page_repositories.setObjectName("page_repositories")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.page_repositories)
        self.verticalLayout.setObjectName("verticalLayout")
        self.listView = QtWidgets.QListView(self.page_repositories)
        self.listView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.listView.setObjectName("listView")
        self.verticalLayout.addWidget(self.listView)
        NewJobWizard.addPage(self.page_repositories)
        self.page_sources = QtWidgets.QWizardPage()
        self.page_sources.setObjectName("page_sources")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.page_sources)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tableView = QtWidgets.QTableView(self.page_sources)
        self.tableView.setAlternatingRowColors(True)
        self.tableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView.setShowGrid(False)
        self.tableView.setWordWrap(False)
        self.tableView.setObjectName("tableView")
        self.tableView.verticalHeader().setVisible(False)
        self.verticalLayout_2.addWidget(self.tableView)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.toolButton_2 = QtWidgets.QToolButton(self.page_sources)
        self.toolButton_2.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.toolButton_2.setObjectName("toolButton_2")
        self.horizontalLayout.addWidget(self.toolButton_2)
        self.toolButton_3 = QtWidgets.QToolButton(self.page_sources)
        self.toolButton_3.setObjectName("toolButton_3")
        self.horizontalLayout.addWidget(self.toolButton_3)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        NewJobWizard.addPage(self.page_sources)
        self.page_schedule = QtWidgets.QWizardPage()
        self.page_schedule.setObjectName("page_schedule")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.page_schedule)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label = QtWidgets.QLabel(self.page_schedule)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_4.addWidget(self.label)
        self.checkBox = QtWidgets.QCheckBox(self.page_schedule)
        self.checkBox.setObjectName("checkBox")
        self.verticalLayout_4.addWidget(self.checkBox)
        self.widget_2 = ScheduleWidget(self.page_schedule)
        self.widget_2.setObjectName("widget_2")
        self.verticalLayout_4.addWidget(self.widget_2)
        NewJobWizard.addPage(self.page_schedule)
        self.wizardPage = QtWidgets.QWizardPage()
        self.wizardPage.setObjectName("wizardPage")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.wizardPage)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_16 = QtWidgets.QLabel(self.wizardPage)
        self.label_16.setWordWrap(True)
        self.label_16.setObjectName("label_16")
        self.verticalLayout_3.addWidget(self.label_16)
        self.plainTextEdit = QtWidgets.QPlainTextEdit(self.wizardPage)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.plainTextEdit.setFont(font)
        self.plainTextEdit.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.plainTextEdit.setReadOnly(True)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout_3.addWidget(self.plainTextEdit)
        NewJobWizard.addPage(self.wizardPage)
        self.action_add_source = QtWidgets.QAction(NewJobWizard)
        self.action_add_source.setObjectName("action_add_source")

        self.retranslateUi(NewJobWizard)
        QtCore.QMetaObject.connectSlotsByName(NewJobWizard)

    def retranslateUi(self, NewJobWizard):
        _translate = QtCore.QCoreApplication.translate
        NewJobWizard.setWindowTitle(_translate("NewJobWizard", "Wizard"))
        self.page_repositories.setTitle(_translate("NewJobWizard", "Repositories"))
        self.page_repositories.setSubTitle(_translate("NewJobWizard", "Select the repositories where data will be saved."))
        self.page_sources.setTitle(_translate("NewJobWizard", "Sources"))
        self.page_sources.setSubTitle(_translate("NewJobWizard", "Select the directories to back up."))
        self.toolButton_2.setText(_translate("NewJobWizard", "..."))
        self.toolButton_3.setText(_translate("NewJobWizard", "..."))
        self.page_schedule.setTitle(_translate("NewJobWizard", "Schedule"))
        self.page_schedule.setSubTitle(_translate("NewJobWizard", "When and how often shall the job run?"))
        self.label.setText(_translate("NewJobWizard", "<html><head/><body><p><span style=\" font-weight:600;\">Note:</span> If you choose to use the global settings with the box below, changes to the global settings made <span style=\" font-style:italic;\">after</span> the job is created will be applied to the job, too. If you simply want to copy the global settings, uncheck the box below.</p></body></html>"))
        self.checkBox.setText(_translate("NewJobWizard", "Use global settings"))
        self.wizardPage.setTitle(_translate("NewJobWizard", "Finished!"))
        self.wizardPage.setSubTitle(_translate("NewJobWizard", "The configuration is finished."))
        self.label_16.setText(_translate("NewJobWizard", "<html><head/><body><p>If needed, you can make modifications to the job configuration below. In case of doubt, simply click continue.</p></body></html>"))
        self.action_add_source.setText(_translate("NewJobWizard", "Add source"))

from qtborg.schedule_widget import ScheduleWidget



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_schedule_widget.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.spinBox = QtWidgets.QSpinBox(Form)
        self.spinBox.setMinimum(1)
        self.spinBox.setMaximum(10000)
        self.spinBox.setObjectName("spinBox")
        self.horizontalLayout.addWidget(self.spinBox)
        self.comboBox_2 = QtWidgets.QComboBox(Form)
        self.comboBox_2.setObjectName("comboBox_2")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.horizontalLayout.addWidget(self.comboBox_2)
        self.formLayout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.comboBox = QtWidgets.QComboBox(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBox.sizePolicy().hasHeightForWidth())
        self.comboBox.setSizePolicy(sizePolicy)
        self.comboBox.setObjectName("comboBox")
        self.comboBox.addItem("")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.comboBox)
        self.checkBox = QtWidgets.QCheckBox(Form)
        self.checkBox.setObjectName("checkBox")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.SpanningRole, self.checkBox)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.spinBox_2 = QtWidgets.QSpinBox(Form)
        self.spinBox_2.setMinimum(1)
        self.spinBox_2.setMaximum(59)
        self.spinBox_2.setObjectName("spinBox_2")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.spinBox_2)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.spinBox_3 = QtWidgets.QSpinBox(Form)
        self.spinBox_3.setMinimum(1)
        self.spinBox_3.setMaximum(24)
        self.spinBox_3.setObjectName("spinBox_3")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.spinBox_3)
        self.label_5 = QtWidgets.QLabel(Form)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.spinBox_4 = QtWidgets.QSpinBox(Form)
        self.spinBox_4.setObjectName("spinBox_4")
        self.horizontalLayout_2.addWidget(self.spinBox_4)
        self.comboBox_3 = QtWidgets.QComboBox(Form)
        self.comboBox_3.setObjectName("comboBox_3")
        self.horizontalLayout_2.addWidget(self.comboBox_3)
        self.formLayout.setLayout(5, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_2)
        self.label_6 = QtWidgets.QLabel(Form)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.spinBox_5 = QtWidgets.QSpinBox(Form)
        self.spinBox_5.setMinimum(1)
        self.spinBox_5.setMaximum(20000)
        self.spinBox_5.setObjectName("spinBox_5")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.spinBox_5)
        self.label_7 = QtWidgets.QLabel(Form)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.spinBox_6 = QtWidgets.QSpinBox(Form)
        self.spinBox_6.setMinimum(1)
        self.spinBox_6.setMaximum(200000)
        self.spinBox_6.setObjectName("spinBox_6")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.spinBox_6)

        self.retranslateUi(Form)
        self.comboBox.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Run every"))
        self.comboBox_2.setItemText(0, _translate("Form", "Minutes"))
        self.comboBox_2.setItemText(1, _translate("Form", "Hours"))
        self.comboBox_2.setItemText(2, _translate("Form", "Days"))
        self.comboBox_2.setItemText(3, _translate("Form", "Weeks"))
        self.comboBox_2.setItemText(4, _translate("Form", "Months"))
        self.comboBox_2.setItemText(5, _translate("Form", "Years"))
        self.label_2.setText(_translate("Form", "Timezone"))
        self.comboBox.setItemText(0, _translate("Form", "Local time"))
        self.checkBox.setText(_translate("Form", "Run job at specific time in the interval"))
        self.label_3.setText(_translate("Form", "Minute"))
        self.label_4.setText(_translate("Form", "Hour"))
        self.label_5.setText(_translate("Form", "Day of week"))
        self.spinBox_4.setToolTip(_translate("Form", "If the interval is longer than one week, select the week to run the job at here."))
        self.label_6.setText(_translate("Form", "Day of month"))
        self.spinBox_5.setToolTip(_translate("Form", "If the interval has less days than the day at which the job is scheduled, the job is scheduled on the last day of the interval."))
        self.label_7.setText(_translate("Form", "Day of year"))
        self.spinBox_6.setToolTip(_translate("Form", "If the interval has less days than the day at which the job is scheduled, the job is scheduled on the last day of the interval."))




########################################################################
# File name: subprocess_backend.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import ast
import asyncio
import base64
import functools
import os
import random
import re
import subprocess

from datetime import datetime

from . import backend


VERSION_RE = re.compile(
    r"^\s*borg\s+"
    r"(?P<major>[0-9]+)\."
    r"(?P<minor>[0-9]+)\."
    r"(?P<release>[0-9]+)\s*$",
)


SIMPLE_ATTR_RE = re.compile(
    r"^(?P<name>[\w \(\)]+):\s+(?P<value>.+)$",
    re.MULTILINE,
)

SIZE_RE = re.compile(
    "[0-9.]+\s+[TGMK]B",
    re.I,
)

CREATE_PROGRESS_RE = re.compile(
    r"^(?P<uncompressed>[0-9.]+\s+[TGMK]?B)\s+O\s+"
    r"(?P<compressed>[0-9.]+\s+[TGMK]?B)\s+C\s+"
    r"(?P<deduplicated>[0-9.]+\s+[TGMK]?B)\s+D\s+"
    r"(?P<nfiles>[0-9]+)\s+N\s+"
    r".*$",
    re.I,
)


SYNC_PROGRESS_RE = re.compile(
    r"^\s*(?P<synced>[0-9]+)% Syncing chunks cache\..+$",
    re.I,
)


DELETE_PROGRESS_RE = re.compile(
    r"^Decrementing references\s+(?P<progress>\d+)%\s*$",
    re.I,
)


LOCK_TIMEOUT_FINGERPRINT_1_0 = (
    b"('Remote Exception (see remote log for the traceback)', 'LockTimeout')"
)

LOCK_TIMEOUT_FINGERPRINT_1_1 = (
    b"Failed to create/acquire the lock",
)

CONNECTION_ERROR_FINGERPRINT = (
    b"Remote: ssh: "
)


def parse_version(s):
    m = VERSION_RE.match(s)
    if m is None:
        raise ValueError("invalid borg version: {!r}".format(s))

    info = m.groupdict()
    return (
        int(info["major"]),
        int(info["minor"]),
        int(info["release"]),
    )


def parse_timestamp(s):
    return datetime.strptime(
        s,
        "%a, %Y-%m-%d %H:%M:%S"
    )


def parse_sizes(s):
    sizes = tuple(SIZE_RE.finditer(s))
    if len(sizes) != 3:
        raise ValueError(
            "failed to parse exactly three sizes "
            "from {!r}".format(s)
        )

    return tuple(
        match.group(0)
        for match in sizes
    )


def format_version(v):
    return ".".join(map(str, v))


def forward_create_progress(cb, s):
    info = {
        "nfiles": 0,
        "uncompressed": 0,
        "compressed": 0,
        "deduplicated": 0,
        "synced": 0,
    }

    parsed = CREATE_PROGRESS_RE.match(s)
    if parsed is not None:
        info = parsed.groupdict()
        info["nfiles"] = int(info["nfiles"])
        info["uncompressed"] = tuple(info["uncompressed"].split())
        info["compressed"] = tuple(info["compressed"].split())
        info["deduplicated"] = tuple(info["deduplicated"].split())
        info["synced"] = 1
        cb(info)
        return

    parsed = SYNC_PROGRESS_RE.match(s)
    if parsed is not None:
        raw_info = parsed.groupdict()
        info["synced"] = int(raw_info["synced"]) / 100
        cb(info)
        return


def forward_delete_progress(cb, s):
    parsed = DELETE_PROGRESS_RE.match(s)
    if parsed is None:
        cb(None)
        return

    info = parsed.groupdict()
    info["progress"] = int(info["progress"]) / 100
    cb(info)


@asyncio.coroutine
def readuntil_either(stream, separators):
    buf = bytearray()
    while True:
        ch = yield from stream.read(1)
        if not ch:
            raise asyncio.streams.IncompleteReadError(
                bytes(buf),
                separators,
            )

        buf.append(ord(ch))
        if ch in separators:
            break

    return bytes(buf)


class LocalSubprocessBackend(backend.Backend):
    SUPPORTED_VERSION_RANGES = [
        ((1, 0, 8), (1, 2, 0)),
    ]

    LIST_FORMAT = (
        "{{"
        "'mode': {mode!r}, "
        "'user': {user!r}, "
        "'group': {group!r}, "
        "'size': {size!r}, "
        "'mtime': {isomtime!r}, "
        "'path': {path!r}, "
        "'uid': {uid!r}, "
        "'gid': {gid!r}, "
        "'type': {type!r}, "
        "'extra': {extra!r}, "
        "'linktarget': {linktarget!r}, "
        "'source': {source!r}, "
        "}}, "
    )

    @staticmethod
    def make_call_id():
        return "call-{}".format(base64.b32encode(
            random.getrandbits(40).to_bytes(5, 'little')
        ).decode("ascii").lower())

    def __init__(self,
                 logger=None, *,  # NOQA
                 path_to_borg="borg"):
        super().__init__(logger=logger)
        self.path_to_borg = path_to_borg

        self._check_version()

    def _get_version(self):
        try:
            s = subprocess.check_output(
                [self.path_to_borg, "--version"],
            ).decode("ascii")
            version = parse_version(s)
        except ValueError as exc:
            raise RuntimeError(
                "could not determine version of borg",
            ) from exc
        return version

    def _check_version(self):
        version = self._get_version()

        self.logger.debug(
            "detected Borg version %s", ".".join(
                map(str, version)
            )
        )

        for ge_version, lt_version in self.SUPPORTED_VERSION_RANGES:
            if ge_version <= version < lt_version:
                return

        raise RuntimeError(
            "cannot use this version of Borg {}. "
            "supported versions are: {}".format(
                format_version(version),
                ", ".join(
                    "{} <= v < {}".format(
                        format_version(v1),
                        format_version(v2),
                    )
                    for v1, v2 in self.SUPPORTED_VERSION_RANGES
                )
            )
        )

    def _raise_common_errors(self, stderr):
        if stderr.startswith(CONNECTION_ERROR_FINGERPRINT):
            raise backend.RepositoryUnreachable(
                stderr.split(b"\n", 1)[0].decode()
            )

        if stderr.startswith(LOCK_TIMEOUT_FINGERPRINT_1_1):
            raise backend.RepositoryLocked()

        if LOCK_TIMEOUT_FINGERPRINT_1_0 in stderr:
            raise backend.RepositoryLocked()

    def _prep_call(self, args, env, context, call_id):
        args = [self.path_to_borg] + args

        if context.dry_run:
            args.insert(2, "-n")

        env = dict(os.environ)
        env.update(env)
        env["LANG"] = "C"

        if context.borg_remote_path is not None:
            env["BORG_REMOTE_PATH"] = context.borg_remote_path

        env["BORG_RELOCATED_REPO_ACCESS_IS_OK"] = "yes"

        self.logger.debug(
            "%s: prepared: %r, with env=%r (passphrase omitted from env)",
            call_id,
            args,
            env,
        )

        if context.passphrase is not None:
            env["BORG_PASSPHRASE"] = context.passphrase

        return args, env

    @asyncio.coroutine
    def _call(self, args, env, call_id):
        self.logger.debug("%s: invoking %s", call_id, args)

        proc = yield from asyncio.create_subprocess_exec(
            *args,
            env=env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        self.logger.debug("%s: started", call_id)

        try:
            _, stderr = yield from proc.communicate()
            self.logger.debug("%s: communicate returned: %r", call_id, stderr)
        except asyncio.CancelledError:
            self.logger.debug("%s: cancelled", call_id)
            if proc.returncode is None:
                self.logger.debug("%s: terminating child", call_id)
                proc.terminate()
                _, stderr = yield from proc.communicate()
                self.logger.debug("%s: communicate returned: %r",
                                  call_id, stderr)
            raise
        finally:
            self.logger.debug("%s: waiting for child", call_id)
            yield from proc.wait()

        self.logger.debug("%s: finished", call_id)
        if proc.returncode != 0:
            self.logger.debug("%s: non-zero exit code", call_id)
            self._raise_common_errors(stderr)

            raise subprocess.CalledProcessError(
                proc.returncode,
                args,
                stderr=stderr,
            )

    @asyncio.coroutine
    def _call_lines(self, args, env, line_callback, call_id):
        self.logger.debug("%s: invoking %s", call_id, args)

        proc = yield from asyncio.create_subprocess_exec(
            *args,
            stderr=subprocess.PIPE,
            stdin=subprocess.DEVNULL,
            env=env,
        )

        self.logger.debug("%s: started", call_id)

        try:
            potential_information = bytearray()
            try:
                while True:
                    line = yield from readuntil_either(
                        proc.stderr,
                        b"\r\n"
                    )
                    self.logger.debug("%s: << %r", call_id, line.strip())
                    # detect and raise errors earily
                    self._raise_common_errors(line)
                    potential_information.extend(line)
                    line_callback(line.decode())
            except asyncio.streams.IncompleteReadError as exc:
                if LOCK_TIMEOUT_FINGERPRINT_1_0 in exc.partial:
                    raise backend.RepositoryLocked() from None
                potential_information.extend(exc.partial)

            yield from proc.wait()
        except:
            self.logger.debug("%s: failure / cancellation", call_id)
            if proc.returncode is None:
                self.logger.debug("%s: terminating child", call_id)
                try:
                    proc.terminate()
                except ProcessLookupError:
                    pass
                self.logger.debug("%s: waiting for child", call_id)
                yield from proc.wait()
            raise

        self.logger.debug("%s: finished", call_id)
        if proc.returncode != 0:
            self.logger.debug("%s: non-zero exit code", call_id)
            self._raise_common_errors(potential_information)
            raise subprocess.CalledProcessError(
                proc.returncode,
                args,
                stderr=potential_information
            )

    def init_repository(self, path, mode, context):
        args = [
            "init",
            "-e", mode.value,
            path,
        ]

        return self._call(args, {}, context)

    def list_archives(self, path, context):
        args = [
            "list",
            path,
        ]

        return self._call(args, {}, context)

    @asyncio.coroutine
    def ping(self, path, context):
        call_id = self.make_call_id()

        args = [
            "info",
            path,
        ]

        env = {}

        args, env = self._prep_call(args, env, context, call_id)

        try:
            yield from self._call(args, env, call_id)
        except backend.RepositoryLocked:
            pass

    @asyncio.coroutine
    def prune_repository(self, path,
                         context,
                         prefix,
                         *,
                         hourly=None,
                         daily=None,
                         weekly=None,
                         monthly=None,
                         yearly=None,
                         keep_within=None):
        call_id = self.make_call_id()

        args = [
            "prune",
            path,
        ]

        if prefix is not None:
            args.extend([
                "--prefix", prefix,
            ])

        for key, value in [
                ("hourly", hourly),
                ("daily", daily),
                ("weekly", weekly),
                ("monthly", monthly),
                ("yearly", yearly)]:
            if value is not None:
                args.extend([
                    "--keep-{}".format(key),
                    str(value),
                ])

        if keep_within is not None:
            args.extend([
                "--keep-within", str(keep_within)
            ])

        args, env = self._prep_call(args, {}, context, call_id)

        return (yield from self._call(args, env, call_id))

    @asyncio.coroutine
    def create_archive(self, path, name, source_paths, context):
        call_id = self.make_call_id()

        args = [
            "create",
        ]
        env = {}

        if context.progress_callback:
            args.append("--progress")

        if context.compression != backend.CompressionMethod.NONE:
            compression_name = context.compression.value
            if context.compression_level is not None:
                compression_name += ","+str(context.compression_level)
            args.extend([
                "-C", compression_name,
            ])

        if context.one_file_system:
            args.append("-x")

        if (context.network_limit_upstream is not None
                or context.network_limit_downstream is not None):
            if context.network_limit_builtin is not None:
                self.logger.warning(
                    "ignoring upstream/downstream limiting and using only borg"
                    " internal ratelimiting"
                )
            else:
                self.logger.warning(
                    "using legacy network limiting is experimental and may "
                    "make borg hang"
                )
                parts = ["-s"]
                if context.network_limit_upstream is not None:
                    parts.extend([
                        "-u", str(context.network_limit_upstream),
                    ])
                if context.network_limit_downstream is not None:
                    parts.extend([
                        "-d", str(context.network_limit_downstream),
                    ])

                env["BORG_RSH"] = "trickle {} ssh".format(" ".join(parts))

        if context.network_limit_builtin is not None:
            args.insert(0, "--remote-ratelimit")
            args.insert(1, str(context.network_limit_builtin))

        args.append("--")
        args.append("::".join([path, name]))
        args.extend(map(str, source_paths))

        args, env = self._prep_call(args, env, context, call_id)

        line_proc = functools.partial(
            forward_create_progress,
            context.progress_callback or (lambda x: None),
        )

        try:
            yield from self._call_lines(args, env, line_proc, call_id)
        except subprocess.CalledProcessError as exc:
            raise RuntimeError(
                "create_archive operation failed: {}".format(
                    exc.stderr.decode()
                )
            ) from None

    def get_archive_info(self, path, name, context):
        args = [
            "info",
            "::".join([path, name])
        ]
        env = {}

        info = self._call(args, env, context).decode("utf-8")

        attrs = {}
        for attr in SIMPLE_ATTR_RE.finditer(info):
            attr_info = attr.groupdict()
            attrs[attr_info["name"]] = attr_info["value"]

        try:
            start_time = parse_timestamp(attrs["Time (start)"])
        except (ValueError, KeyError):
            self.logger.warning("failed to parse timestamp", exc_info=True)
            start_time = None

        try:
            end_time = parse_timestamp(attrs["Time (end)"])
        except (KeyError, ValueError):
            self.logger.warning("failed to parse timestamp", exc_info=True)
            end_time = None

        try:
            nfiles = int(attrs["Number of files"])
        except (KeyError, ValueError):
            self.logger.warning("failed to parse number of files",
                                exc_info=True)
            nfiles = None

        try:
            uncompressed_size, compressed_size, deduplicated_size = \
                parse_sizes(attrs["This archive"])
        except (KeyError, ValueError):
            self.logger.warning("failed to parse sizes",
                                exc_info=True)
            uncompressed_size = None
            compressed_size = None
            deduplicated_size = None

        return backend.ArchiveInfo(
            name=attrs.get("Name"),
            fingerprint=attrs.get("Fingerprint"),
            hostname=attrs.get("Hostname"),
            username=attrs.get("Username"),
            start_time=start_time,
            end_time=end_time,
            commandline=attrs.get("Command line"),
            nfiles=nfiles,
            uncompressed_size=uncompressed_size,
            compressed_size=compressed_size,
            deduplicated_size=deduplicated_size,
        )

    def list_archive_contents(self, path, name, context, *, prefix=None):
        args = [
            "list",
            "::".join([path, name]),
            "--list-format", self.LIST_FORMAT,
        ]

        if prefix is not None:
            args.extend([
                "--prefix", prefix,
            ])

        env = {}

        items = ast.literal_eval(
            "[{}]".format(self._call(args, env, context).decode())
        )

        for item in items:
            try:
                item["mtime"] = parse_timestamp(item["mtime"])
            except ValueError:
                self.logger.warning("failed to parse timestamp", exc_info=True)
                item["mtime"] = None

        return items

    @asyncio.coroutine
    def delete_archive(self, path, name, context):
        args = [
            "delete",
            "::".join([path, name]),
        ]

        env = {}

        if context.progress_callback:
            args.append("--progress")

        args, env = self._prep_call(args, env, context)

        line_proc = functools.partial(
            forward_delete_progress,
            context.progress_callback or (lambda x: None),
        )

        yield from self._call_lines(
            args,
            env,
            functools.partial(
                forward_delete_progress,
                context.progress_callback,
            )
        )

    def delete_archive_cache(self, path, name, context):
        args = [
            "delete",
            "::".join([path, name]),
            "--cache-only",
        ]

        env = {}

        if context.progress_callback:
            args.append("--progress")

        if context.progress_callback:
            return self._call_lines(
                args,
                env,
                context,
                functools.partial(
                    forward_delete_progress,
                    context.progress_callback,
                )
            )
        else:
            return self._call(args, env, context)



########################################################################
# File name: wctsleep.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import logging

from datetime import datetime

if not hasattr(asyncio, "ensure_future"):
    asyncio.ensure_future = asyncio.async


MAX_SLACK = 60


logger = logging.getLogger(__name__)


class WallClockTimeSleepImpl:
    def __init__(self, max_slack=MAX_SLACK):
        super().__init__()
        self._futures = []
        self._wakeup_event = asyncio.Event()
        self._main_task = None
        self._max_slack = max_slack

    @property
    def max_slack(self):
        return self._max_slack

    @max_slack.setter
    def max_slack(self, new_value):
        self._max_slack = new_value
        self._wakeup_event.set()

    @asyncio.coroutine
    def cancel_and_wait(self):
        if self._main_task is None:
            return

        for _, fut in self._futures:
            fut.cancel()
        self._wakeup_event.set()
        yield from self._main_task

    def _main_done(self, task):
        pass

    def _cleanup(self):
        self._futures = sorted(
            (
                (ts, fut)
                for ts, fut in self._futures
                if not fut.done()
            ),
            key=lambda x: x[0],
        )

    @asyncio.coroutine
    def _main(self):
        # self._cleanup always returns None
        while self._cleanup() or self._futures:
            now = datetime.utcnow()
            next_event, _ = self._futures[0]
            wait_time = (next_event - now).total_seconds()
            if wait_time >= 0:
                wait_time = min(wait_time, self._max_slack)
                logger.debug("sleeping for %.1f seconds",
                             wait_time)
                try:
                    yield from asyncio.wait_for(
                        self._wakeup_event.wait(),
                        timeout=wait_time
                    )
                except asyncio.TimeoutError:
                    logger.debug("timeout passed")
                else:
                    self._wakeup_event.clear()
                    logger.debug("woke up from event")
                continue

            for i, (timestamp, fut) in enumerate(self._futures):
                if timestamp > now:
                    break
                fut.set_result(None)

    def _notify(self):
        # start task if its not running
        if self._main_task is None or self._main_task.done():
            self._main_task = asyncio.ensure_future(
                self._main()
            )
            self._main_task.add_done_callback(self._main_done)
            self._wakeup_event.clear()
        else:
            self._wakeup_event.set()

    def sleep_until(self, dt):
        fut = asyncio.Future()
        self._futures.append((dt, fut))
        self._notify()
        return fut



########################################################################
# File name: utils.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
def print_table(header, rows):
    col_max_widths = [len(hdr) for hdr in header]

    for row in rows:
        for col, cell in enumerate(row):
            col_max_widths[col] = max(
                len(cell),
                col_max_widths[col],
            )

    fmt = " | ".join(
        "{{:<{w}s}}".format(
            w=width
        )
        for width in col_max_widths,
    )

    print(fmt.format(*("="*w for w in col_max_widths)))
    print(fmt.format(*header))
    print(fmt.format(*("-"*w for w in col_max_widths)))

    for row in rows:
        print(fmt.format(*row))



########################################################################
# File name: cli.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import asyncio
import inspect
import logging
import os
import pathlib
import pprint
import sys

from datetime import datetime

import xdg.BaseDirectory

from .. import utils
from . import core, status, protocol


def cmd_run(loop, args):
    logger = logging.getLogger("cuteborg.scheduler")

    scheduler = core.Scheduler(loop, args.socket_path)

    server = loop.run_until_complete(scheduler.check_server_socket())
    sock = scheduler.create_server_socket()

    if args.daemonize:
        logger.debug("forking into background")

        try:
            pid = os.fork()
        except OSError:
            logger.error("failed to fork")
            return

        if pid != 0:
            # parent
            return 0

        # child
        # start new session
        os.setsid()
        try:
            pid = os.fork()
        except OSError:
            logger.error("second fork failed")
            return 1

        if pid != 0:
            # parent -> exit
            return 0

        logger.debug("double-forked into background")

    asyncio.get_event_loop().close()
    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)

    loop.run_until_complete(scheduler.main(sock=sock, loop=loop))

    sys.stdout.flush()
    sys.stderr.flush()


def _format_value(value):
    if isinstance(value, datetime):
        return str(value)
    return repr(value)


def _format_dict(d):
    return " ".join(
        "{}={}".format(key, _format_value(value))
        for key, value
        in sorted(d.items())
    )


def _format_status_row(item):
    reason_map = {
        "waiting_for_repository_lock": "repolock",
        "removable_device": "device",
        "sleep_until": "scheduled",
    }

    try:
        args = item["args"]
    except KeyError:
        args = str(item["raw_args"])
    else:
        args = _format_dict(args)

    state = "??"
    if item["running"] and not item.get("blocking"):
        state = "run"
    elif item.get("error"):
        state = "error"
    elif item.get("blocking"):
        state = "wait"

    status_parts = []
    try:
        blocking_info = item["blocking"]
    except KeyError:
        pass
    else:
        status_parts.append(
            reason_map.get(blocking_info["reason"],
                           blocking_info["reason"])
        )

        try:
            struct_info = blocking_info["struct_info"]
        except KeyError:
            status_parts.append(
                str(blocking_info["raw_info"])
            )
        else:
            status_parts.append(
                _format_dict(struct_info)
            )

    try:
        error_info = item["error"]
    except KeyError:
        pass
    else:
        status_parts.append("error since {}: {}".format(
            error_info["since"],
            error_info["message"],
        ))

    try:
        progress_info = item["progress"]
    except KeyError:
        pass
    else:
        status_parts.append("progress")
        status_parts.append(
            _format_dict(progress_info)
        )

    return (
        item["type"],
        args,
        state,
        " ".join(status_parts),
    )


@asyncio.coroutine
def cmd_status(loop, args):
    data = yield from status.get_status(loop, args.socket_path)

    logging.getLogger("cuteborg.scheduler").debug(
        "\n%s",
        pprint.pformat(data),
    )

    headers = [
        "type",
        "args",
        "state",
        "status",
    ]

    rows = list(map(_format_status_row, data["jobs"]))

    print("jobs: {}".format(len(data["jobs"])))
    utils.print_table(headers, rows)


@asyncio.coroutine
def cmd_reload(loop, args):
    logger = logging.getLogger("cuteborg.scheduler")

    try:
        endpoint = yield from asyncio.wait_for(
            protocol.test_and_get_socket(
                loop,
                logger,
                args.socket_path,
            ),
            timeout=5
        )
    except (asyncio.TimeoutError, FileNotFoundError, ConnectionError) as exc:
        msg = str(exc)
        if isinstance(exc, asyncio.TimeoutError):
            msg = "timeout"
        logger.error("failed to connect to scheduler at %r: %s",
                     str(args.socket_path),
                     msg)
        raise RuntimeError(
            "failed to connect to scheduler"
        )

    response, data = yield from endpoint.request(
        protocol.ToplevelCommand.RESCHEDULE
    )

    if response == protocol.ToplevelCommand.OKAY:
        logger.info("reload triggered successfully")
        return 0
    else:
        logger.error("could not trigger reload")
        return 2


@asyncio.coroutine
def cmd_stop(loop, args):
    logger = logging.getLogger("cuteborg.scheduler")

    try:
        endpoint = yield from asyncio.wait_for(
            protocol.test_and_get_socket(
                loop,
                logger,
                args.socket_path,
            ),
            timeout=5
        )
    except (asyncio.TimeoutError, FileNotFoundError, ConnectionError) as exc:
        msg = str(exc)
        if isinstance(exc, asyncio.TimeoutError):
            msg = "timeout"
        logger.error("failed to connect to scheduler at %r: %s",
                     str(args.socket_path),
                     msg)
        raise RuntimeError(
            "failed to connect to scheduler"
        )

    response, data = yield from endpoint.request(
        protocol.ToplevelCommand.EXIT
    )

    if response != protocol.ToplevelCommand.OKAY:
        logger.error("could not stop scheduler")
        return 2

    logger.info("scheduler stopping")

    try:
        yield from endpoint.protocol.breakdown_future
    except OSError:
        pass
    logger.info("scheduler stopped")


@asyncio.coroutine
def cmd_force_archive(loop, args):
    logger = logging.getLogger("cuteborg.scheduler")

    try:
        endpoint = yield from asyncio.wait_for(
            protocol.test_and_get_socket(
                loop,
                logger,
                args.socket_path,
            ),
            timeout=5
        )
    except (asyncio.TimeoutError, FileNotFoundError, ConnectionError) as exc:
        msg = str(exc)
        if isinstance(exc, asyncio.TimeoutError):
            msg = "timeout"
        logger.error("failed to connect to scheduler at %r: %s",
                     str(args.socket_path),
                     msg)
        raise RuntimeError(
            "failed to connect to scheduler"
        )

    request = {
        "command": "force-archive",
        "force-archive": {
            "name": args.job,
            "only": list(set(args.only)),
        }
    }

    response, data = yield from endpoint.request(
        protocol.ToplevelCommand.EXTENDED_REQUEST,
        request,
    )

    print(response, data)


@asyncio.coroutine
def cmd_force_prune(loop, args):
    logger = logging.getLogger("cuteborg.scheduler")

    try:
        endpoint = yield from asyncio.wait_for(
            protocol.test_and_get_socket(
                loop,
                logger,
                args.socket_path,
            ),
            timeout=5
        )
    except (asyncio.TimeoutError, FileNotFoundError, ConnectionError) as exc:
        msg = str(exc)
        if isinstance(exc, asyncio.TimeoutError):
            msg = "timeout"
        logger.error("failed to connect to scheduler at %r: %s",
                     str(args.socket_path),
                     msg)
        raise RuntimeError(
            "failed to connect to scheduler"
        )

    request = {
        "command": "force-prune",
        "force-prune": {
            "repositories": list(set(args.repositories)),
        }
    }

    response, data = yield from endpoint.request(
        protocol.ToplevelCommand.EXTENDED_REQUEST,
        request,
    )

    print(response, data)


def autoloop(loop, f, *args, **kwargs):
    if asyncio.iscoroutinefunction(f):
        return loop.run_until_complete(f(*args, **kwargs))
    else:
        return f(*args, **kwargs)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Increase verbosity (up to -vvv)"
    )

    parser.add_argument(
        "--socket-path",
        default=None,
        type=pathlib.Path,
        help="Override the socket path "
        "(defaults to a user-specific socket)"
    )

    subparsers = parser.add_subparsers()

    subparser = subparsers.add_parser(
        "run",
        help="Execute the scheduler (runs in foreground)"
    )
    subparser.set_defaults(func=cmd_run)
    subparser.add_argument(
        "-d", "--daemonize",
        action="store_true",
        default=False,
        help="Daemonize after setup instead of running in foreground",
    )

    subparser = subparsers.add_parser(
        "status",
        help="Connect to a running scheduler and show the status"
    )
    subparser.set_defaults(func=cmd_status)

    subparser = subparsers.add_parser(
        "reload",
        help="Connect to a running scheduler and ask it to schedule a "
        "reload and reschedule"
    )
    subparser.set_defaults(func=cmd_reload)

    subparser = subparsers.add_parser(
        "stop",
        help="Connect to a running scheduler and ask it to stop"
    )
    subparser.set_defaults(func=cmd_stop)

    subparser = subparsers.add_parser(
        "force-archive",
        help="Force an archive job to run",
    )
    subparser.add_argument(
        "job",
        help="Name of the job",
    )
    subparser.add_argument(
        "--only",
        action="append",
        default=[],
        help="Only run the job on this repository. May be given multiple times."
    )
    subparser.set_defaults(func=cmd_force_archive)

    subparser = subparsers.add_parser(
        "force-prune",
        help="Force a prune job to run",
    )
    subparser.add_argument(
        "repositories",
        metavar="REPO-ID",
        nargs="+",
        help="ID of the repositories to prune",
    )
    subparser.set_defaults(func=cmd_force_prune)

    args = parser.parse_args()

    if not hasattr(args, "func"):
        print("select a subcommand (use --help)", file=sys.stderr)
        sys.exit(1)

    logging.basicConfig(
        level={
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
        }.get(args.verbosity, logging.DEBUG),
    )

    if args.socket_path is None:
        args.socket_path = pathlib.Path(
            xdg.BaseDirectory.get_runtime_dir(strict=False)
        ) / "cuteborg" / "scheduler.sock"

    loop = asyncio.get_event_loop()
    try:
        sys.exit(autoloop(loop, args.func, loop, args) or 0)
    finally:
        loop.close()



########################################################################
# File name: devices.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import json
import threading
import pathlib
import pprint
import subprocess

import dbus


_locals = threading.local()


class EncryptedDeviceWithoutPassphrase(Exception):
    pass


@asyncio.coroutine
def lsblk():
    args = [
        "lsblk",
        "--json",
        "-o",
        "NAME,FSTYPE,UUID,MOUNTPOINT,KNAME",
    ]

    proc = yield from asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    stdout, stderr = yield from proc.communicate()
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(
            proc.returncode,
            args,
            stderr=stderr,
        )

    info = json.loads(stdout.decode())

    return info


def lsblk_find_device(devs, uuid):
    for dev in devs:
        if dev["uuid"] == uuid:
            return dev
        try:
            subtree = dev["children"]
        except KeyError:
            pass
        else:
            found_in_subtree = lsblk_find_device(
                subtree,
                uuid,
            )
            if found_in_subtree is not None:
                return found_in_subtree


def _get_threadlocal_bus():
    try:
        return _locals.bus
    except AttributeError:
        _locals.bus = dbus.SystemBus()
        return _locals.bus


def _get_udisks_objects(bus):
    root = bus.get_object(
        "org.freedesktop.UDisks2",
        "/org/freedesktop/UDisks2",
    )

    root_object_manager = dbus.Interface(
        root,
        "org.freedesktop.DBus.ObjectManager",
    )

    return root_object_manager.GetManagedObjects()


def _unlock_device_via_dbus(logger, uuid, crypto_passphrase):
    bus = _get_threadlocal_bus()

    for obj_path in _get_udisks_objects(bus):
        if not str(obj_path).startswith(
                "/org/freedesktop/UDisks2/block_devices/"):
            continue

        dev_obj = bus.get_object(
            "org.freedesktop.UDisks2",
            obj_path,
        )

        dev_props = dbus.Interface(
            dev_obj,
            "org.freedesktop.DBus.Properties",
        )

        dev_uuid = dev_props.Get(
            "org.freedesktop.UDisks2.Block",
            "IdUUID",
        )

        if str(dev_uuid) == uuid:
            dev_encrypted = dbus.Interface(
                dev_obj,
                "org.freedesktop.UDisks2.Encrypted",
            )

            cleartext_obj = bus.get_object(
                "org.freedesktop.UDisks2",
                dev_encrypted.Unlock(crypto_passphrase, []),
            )
            cleartext_dev_props = dbus.Interface(
                cleartext_obj,
                "org.freedesktop.DBus.Properties",
            )

            return bytes(cleartext_dev_props.Get(
                "org.freedesktop.UDisks2.Block",
                "PreferredDevice",
            ))[:-1]  # strip trailing NUL


def _mount_device_via_dbus(logger, uuid):
    bus = _get_threadlocal_bus()

    for obj_path in _get_udisks_objects(bus):
        if not str(obj_path).startswith(
                "/org/freedesktop/UDisks2/block_devices/"):
            continue

        dev_obj = bus.get_object(
            "org.freedesktop.UDisks2",
            obj_path,
        )

        dev_props = dbus.Interface(
            dev_obj,
            "org.freedesktop.DBus.Properties",
        )

        dev_uuid = dev_props.Get(
            "org.freedesktop.UDisks2.Block",
            "IdUUID",
        )

        if str(dev_uuid) == uuid:
            dev_fs = dbus.Interface(
                dev_obj,
                "org.freedesktop.UDisks2.Filesystem",
            )

            return str(dev_fs.Mount([]))


@asyncio.coroutine
def unlock_device_via_dbus(logger, uuid, crypto_passphrase):
    loop = asyncio.get_event_loop()
    return (yield from loop.run_in_executor(
        None,
        _unlock_device_via_dbus,
        logger,
        uuid,
        crypto_passphrase,
    ))


@asyncio.coroutine
def mount_device_via_dbus(logger, uuid):
    loop = asyncio.get_event_loop()
    return (yield from loop.run_in_executor(
        None,
        _mount_device_via_dbus,
        logger,
        uuid,
    ))


@asyncio.coroutine
def wait_for_uuid(logger, uuid, poll_interval):
    p = pathlib.Path("/dev/disk/by-uuid") / uuid
    while True:
        try:
            p = p.resolve()
        except FileNotFoundError:
            pass
        else:
            break

        logger.debug("%s is not present yet ... "
                     "waiting for %.0f seconds",
                     p,
                     poll_interval)
        yield from asyncio.sleep(poll_interval)
    logger.debug("%s appeared", p)


@asyncio.coroutine
def wait_for_mounted(logger, uuid, poll_interval,
                     crypto_passphrase=None):
    while True:
        info = yield from lsblk()
        dev_info = lsblk_find_device(info["blockdevices"], uuid)
        if dev_info is None:
            logger.debug(
                "device %r not there yet ... continuing to wait",
                uuid
            )
            yield from asyncio.sleep(poll_interval)
            continue

        logger.debug(
            "device %r is there, let’s take a look at it",
            uuid
        )

        if dev_info["fstype"] == "crypto_LUKS":
            logger.debug(
                "device %r is encrypted ...",
                uuid,
            )
            if dev_info.get("children", []):
                logger.debug(
                    "but %r appears to be unlocked already :-)",
                    uuid,
                )
                dev_info = dev_info["children"][0]
            else:
                if crypto_passphrase is None:
                    logger.error(
                        "%r is locked, and we do not have a password to try",
                        uuid,
                    )
                    raise EncryptedDeviceWithoutPassphrase(
                        "no passphrase for encrypted device"
                    )

                logger.debug(
                    "%r is locked, and we have a passphrase. "
                    "let’s give it a shot.",
                    uuid,
                )
                yield from unlock_device_via_dbus(
                    logger,
                    dev_info["uuid"],
                    crypto_passphrase,
                )
                logger.debug(
                    "unlocking of %r appears to have worked",
                    uuid,
                )
                # rescan now
                continue

        if dev_info["mountpoint"] is not None:
            logger.debug(
                "device %r is mounted at %r",
                uuid,
                dev_info["mountpoint"]
            )
            return dev_info["mountpoint"]

        logger.debug(
            "attempting to mount %r via dbus",
            uuid,
        )

        return (yield from mount_device_via_dbus(
            logger,
            dev_info["uuid"],
        ))

        yield from asyncio.sleep(poll_interval)
        continue



########################################################################
# File name: config.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import enum
import pathlib
import tempfile

import pytz

import toml

import xdg.BaseDirectory

from .backend import EncryptionMode, CompressionMethod


class IntervalUnit(enum.Enum):
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    YEAR = "year"


JOB_DEFAULTS = {
    "one_file_system": True,
}


def load_config():
    cfg_path = pathlib.Path(
        xdg.BaseDirectory.xdg_config_home
    ) / "cuteborg" / "config.toml"
    try:
        with cfg_path.open("r") as f:
            return toml.load(f)
    except OSError:
        return {}


class ConfigElementBase(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def load_from_raw(self, raw):
        pass

    @abc.abstractmethod
    def to_raw(self):
        return {}


class PruneConfig(ConfigElementBase):
    hourly = None
    daily = None
    weekly = None
    monthly = None
    yearly = None
    keep_within = None

    schedule = None

    def load_from_raw(self, raw):
        self.hourly = raw.get("hourly")
        self.daily = raw.get("daily")
        self.weekly = raw.get("weekly")
        self.monthly = raw.get("monthly")
        self.yearly = raw.get("yearly")
        self.keep_within = raw.get("keep_within")

        try:
            schedule_cfg = raw["schedule"]
        except KeyError:
            pass
        else:
            self.schedule = ScheduleConfig.from_raw(schedule_cfg)

    def _set_raw_attr(self, raw, attr):
        value = getattr(self, attr)
        if value is not None:
            raw[attr] = value

    def to_raw(self):
        raw = super().to_raw()
        self._set_raw_attr(raw, "hourly")
        self._set_raw_attr(raw, "daily")
        self._set_raw_attr(raw, "weekly")
        self._set_raw_attr(raw, "monthly")
        self._set_raw_attr(raw, "yearly")
        self._set_raw_attr(raw, "keep_within")

        if self.schedule is not None:
            raw["schedule"] = self.schedule.to_raw()

        return raw

    @classmethod
    def from_raw(cls, raw):
        self = cls()
        self.load_from_raw(raw)
        return self

    def to_kwargs(self):
        return {
            "hourly": self.hourly,
            "daily": self.daily,
            "weekly": self.weekly,
            "monthly": self.monthly,
            "yearly": self.yearly,
            "keep_within": self.keep_within,
        }


class RepositoryConfig(ConfigElementBase):
    id_ = None
    name = None
    encryption_mode = None
    encryption_passphrase = None
    prune = None

    @abc.abstractmethod
    def load_from_raw(self, raw):
        self.id_ = raw["id"]
        self.name = raw.get("name", self.id_)

        encryption_cfg = raw["encryption"]

        self.encryption_mode = EncryptionMode(encryption_cfg["mode"])
        self.encryption_passphrase = encryption_cfg.get("passphrase")

        try:
            prune_cfg = raw["prune"]
        except KeyError:
            self.prune = None
        else:
            self.prune = PruneConfig.from_raw(prune_cfg)

    @abc.abstractmethod
    def to_raw(self):
        raw = super().to_raw()
        raw["id"] = self.id_
        if self.id_ != self.name:
            raw["name"] = self.name

        encryption_cfg = raw.setdefault("encryption", {})
        encryption_cfg["mode"] = self.encryption_mode.value
        if self.encryption_passphrase is not None:
            encryption_cfg["passphrase"] = self.encryption_passphrase

        if self.prune is not None:
            prune_cfg = self.prune.to_raw()
            raw["prune"] = prune_cfg

        return raw

    @classmethod
    def from_raw(self, raw):
        try:
            type_ = raw["type"]
        except KeyError:
            raise ValueError("repository definition lacks type")

        if type_ == "local":
            result = LocalRepositoryConfig()
        elif type_ == "remote":
            result = RemoteRepositoryConfig()
        else:
            raise ValueError("unknown repository type: {!r}".format(
                type_,
            ))

        result.load_from_raw(raw)
        return result

    def setup_context(self, ctx):
        if self.encryption_mode != EncryptionMode.NONE:
            ctx.passphrase = self.encryption_passphrase


class LocalRepositoryConfig(RepositoryConfig):
    path = None
    removable_device_uuid = None
    crypto_enabled = False
    crypto_passphrase = None

    def load_from_raw(self, raw):
        if raw["type"] != "local":
            raise ValueError("not a local repository config")

        super().load_from_raw(raw)
        local_cfg = raw["local"]

        self.path = local_cfg["path"]

        try:
            removable_cfg = local_cfg["removable"]
        except KeyError:
            self.removable_device_uuid = None
        else:
            self.removable_device_uuid = removable_cfg["device_uuid"]

        try:
            crypto_cfg = local_cfg["crypto"]
        except KeyError:
            self.crypto_enabled = False
            self.crypto_passphrase = None
        else:
            self.crypto_enabled = crypto_cfg.get("enabled", True)
            self.crypto_passphrase = crypto_cfg.get("passphrase")

    def to_raw(self):
        raw = super().to_raw()
        raw.setdefault("local", {})
        raw["local"]["path"] = self.path

        if self.removable_device_uuid is not None:
            removable_cfg = raw["local"].setdefault("removable")
            removable_cfg["device_uuid"] = self.removable_device_uuid

        if self.crypto_enabled:
            crypto_cfg = raw["local"].setdefault("crypto")
            crypto_cfg["enabled"] = self.crypto_enabled
            crypto_cfg["passphrase"] = self.crypto_passphrase

        return raw

    def make_repository_path(self):
        return self.path

    def setup_context(self, ctx):
        super().setup_context(ctx)


class RemoteRepositoryConfig(RepositoryConfig):
    path = None
    borg_path = None
    host = None
    user = None

    def load_from_raw(self, raw):
        if raw["type"] != "remote":
            raise ValueError("not a remote repository config")

        super().load_from_raw(raw)
        remote_cfg = raw["remote"]

        self.path = remote_cfg["path"]
        self.host = remote_cfg["host"]
        self.user = remote_cfg.get("user")
        self.borg_path = remote_cfg.get("borg_path")
        self.ratelimit = int(remote_cfg.get("ratelimit", 0))
        if self.ratelimit < 0:
            self.ratelimit = None

    def to_raw(self):
        raw = super().to_raw()

        remote_cfg = raw.setdefault("remote", {})
        remote_cfg["path"] = self.path
        remote_cfg["host"] = self.host

        if self.user is not None:
            remote_cfg["user"] = self.user

        if self.borg_path is not None:
            remote_cfg["borg_path"] = self.borg_path

        if self.ratelimit is not None:
            remote_cfg["ratelimit"] = self.ratelimit

        return raw

    def make_repository_path(self):
        parts = ["ssh://"]
        if self.user is not None:
            parts.append(self.user)
            parts.append("@")
        parts.append(self.host)
        parts.append(self.path)

        return "".join(parts)

    def setup_context(self, ctx):
        super().setup_context(ctx)
        ctx.borg_remote_path = self.borg_path
        ctx.network_limit_builtin = self.ratelimit


class ScheduleConfig(ConfigElementBase):
    interval_unit = None
    interval_step = None
    timezone = None

    def load_from_raw(self, raw):
        super().load_from_raw(raw)
        self.interval_unit = IntervalUnit(raw["interval_unit"])
        self.interval_step = raw["interval_step"]
        self.timezone = raw.get("timezone", None)
        if self.timezone is not None:
            self.timezone = pytz.timezone(self.timezone)

    def to_raw(self):
        raw = super().to_raw()
        raw["interval_unit"] = self.interval_unit.value
        raw["interval_level"] = self.interval_level
        if self.timezone is not None:
            raw["timezone"] = self.timezone.zone

        return raw

    @classmethod
    def from_raw(cls, raw):
        self = cls()
        self.load_from_raw(raw)
        return self

    def __repr__(self):
        return (
            "<{}.{} "
            "interval_unit={} "
            "interval_step={} "
            "timezone={!r} "
            "at 0x{:x}>".format(
                type(self).__module__,
                type(self).__qualname__,
                self.interval_unit,
                self.interval_step,
                self.timezone,
                id(self),
            )
        )


class JobConfig(ConfigElementBase):
    name = None
    compression_method = None
    compression_level = None
    schedule = None

    def __init__(self):
        super().__init__()
        self.repositories = []

    def load_from_raw(self, raw, repositories):
        super().load_from_raw(raw)
        self.name = raw["name"]
        self.repositories = [
            repositories[repo_id]
            for repo_id in raw["repositories"]
        ]
        self.sources = list(raw["sources"])

        try:
            schedule_cfg = raw["schedule"]
        except KeyError:
            self.schedule = None
        else:
            self.schedule = ScheduleConfig.from_raw(schedule_cfg)

        try:
            compression_cfg = raw["compression"]
        except KeyError:
            self.compression_method = None
            self.compression_level = None
        else:
            self.compression_method = CompressionMethod(
                compression_cfg["method"]
            )
            self.compression_level = compression_cfg.get("level")

    def to_raw(self):
        raw = super().to_raw()
        raw["name"] = self.name
        raw["repositories"] = [
            repo.id_
            for repo in self.repositories
        ]
        raw["sources"] = list(self.sources)

        if self.compression_method is not None:
            compression_cfg = raw.setdefault("compression", {})
            compression_cfg["method"] = self.compression_method.value
            if self.compression_level is None:
                compression_cfg["level"] = self.compression_level

        if self.schedule is not None:
            raw["schedule"] = self.schedule.to_raw()

    @classmethod
    def from_raw(self, raw, repositories):
        try:
            type_ = raw["type"]
        except KeyError:
            raise ValueError("job definition lacks type")

        if type_ == "simple":
            result = JobConfig()
        else:
            raise ValueError("unknown job type: {!r}".format(
                type_
            ))

        result.load_from_raw(raw, repositories)
        return result

    def setup_context(self, ctx):
        ctx.compression_method = self.compression_method
        ctx.compression_level = self.compression_level


class Config(ConfigElementBase):
    poll_interval = 60
    max_slack = 60

    def __init__(self):
        super().__init__()
        self.repositories = {}
        self.jobs = {}

    def load_from_raw(self, raw):
        try:
            schedule_cfg = raw["schedule"]
        except KeyError:
            self.schedule = None
        else:
            self.schedule = ScheduleConfig.from_raw(schedule_cfg)

        for i, repo_raw in enumerate(raw["repository"]):
            if "id" not in repo_raw:
                raise ValueError(
                    "repository #{} (0-based) has no id".format(i)
                )

            try:
                repo = RepositoryConfig.from_raw(repo_raw)
            except (KeyError, ValueError) as exc:
                raise ValueError(
                    "failed to load repository #{} (0-based)".format(i)
                ) from exc

            self.repositories[repo.id_] = repo

        for i, job_raw in enumerate(raw["job"]):
            if "name" not in job_raw:
                raise ValueError(
                    "job #{} (0-based) has no name".format(i)
                )

            try:
                job = JobConfig.from_raw(job_raw, self.repositories)
            except (KeyError, ValueError) as exc:
                raise ValueError(
                    "failed to load job #{} (0-based)".format(i)
                ) from exc

            self.jobs[job.name] = job

        self.poll_interval = raw.get("poll_interval", 60)
        self.max_slack = raw.get("max_slack", 60)

    def to_raw(self):
        raw = super().to_raw()
        raw["repository"] = [
            repo.to_raw()
            for repo in self.repositories.values()
        ]

        raw["job"] = [
            job.to_raw()
            for job in self.jobs.values()
        ]

        raw["poll_interval"] = self.poll_interval
        raw["max_slack"] = self.max_slack

    @classmethod
    def from_raw(cls, raw):
        self = cls()
        self.load_from_raw(raw)
        return self


def toml_to_file(data, path):
    with tempfile.NamedTemporaryFile(
            "w",
            dir=str(path.parent),
            delete=False) as temp:
        tempfile_path = path.parent / temp.name
        try:
            toml.dump(data, temp)
        except:
            try:
                tempfile_path.unlink()
            except OSError:
                # ignore, we want to re-raise the original problem instead
                pass
            raise

        tempfile_path.rename(path)


def save_config(cfg):
    cfg_path = pathlib.Path(
        xdg.BaseDirectory.xdg_config_home
    ) / "cuteborg" / "config.toml"
    toml_to_file(cfg, cfg_path)


def find_repository_by_id(repositories, id_):
    for i, repo in enumerate(repositories):
        if repo["id"] == id_:
            return i
    return None



########################################################################
# File name: backend.py
# This file is part of: cuteborg
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import collections
import enum
import logging


class RepositoryLocked(Exception):
    def __init__(self):
        super().__init__(
            "the repository is locked",
        )


class RepositoryUnreachable(ConnectionError):
    def __init__(self, extra=None):
        msg = "the repository is unreachable"
        if extra is not None:
            msg = "{}: {}".format(msg, extra)
        super().__init__(msg)


class EncryptionMode(enum.Enum):
    NONE = "none"
    KEYFILE = "keyfile"
    REPOKEY = "repokey"


class CompressionMethod(enum.Enum):
    NONE = "none"
    LZ4 = "lz4"
    LZMA = "lzma"
    ZLIB = "zlib"


ArchiveInfo = collections.namedtuple(
    "ArchiveInfo",
    [
        "name",
        "fingerprint",
        "hostname",
        "username",
        "start_time",
        "end_time",
        "commandline",
        "nfiles",
        "uncompressed_size",
        "compressed_size",
        "deduplicated_size",
    ]
)


class Context:
    network_limit_upstream = None
    network_limit_downstream = None
    network_limit_builtin = None
    one_file_system = False
    compression = CompressionMethod.NONE
    compression_level = None
    passphrase = None
    dry_run = False
    progress_callback = None
    borg_remote_path = None


class Backend(metaclass=abc.ABCMeta):
    """
    Backend which interfaces with Borg.
    """

    def __init__(self, logger=None, **kwargs):
        super().__init__(**kwargs)
        if logger is None:
            logger = logging.getLogger(".".join([type(self).__module__,
                                                 type(self).__qualname__]))
        self.logger = logger

    @abc.abstractmethod
    def init_repository(self, path, mode, context):
        pass

    @abc.abstractmethod
    def list_archives(self, path, context):
        pass

    @abc.abstractmethod
    def prune_repository(self, path,
                         context,
                         prefix=None,
                         *,
                         hourly=None,
                         daily=None,
                         weekly=None,
                         monthly=None,
                         yearly=None,
                         keep_within=None):
        pass

    @abc.abstractmethod
    def create_archive(self, path, name, source_paths, context):
        pass

    @abc.abstractmethod
    def get_archive_info(self, path, name, context):
        pass

    @abc.abstractmethod
    def list_archive_contents(self, path, name, context):
        pass

    @abc.abstractmethod
    def delete_archive(self, path, name, context):
        pass



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/trayicon/ui_main.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_StatusWindow(object):
    def setupUi(self, StatusWindow):
        StatusWindow.setObjectName("StatusWindow")
        StatusWindow.resize(851, 398)
        self.centralwidget = QtWidgets.QWidget(StatusWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.status_box = QtWidgets.QWidget(self.centralwidget)
        self.status_box.setObjectName("status_box")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.status_box)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.status_icon = QtWidgets.QLabel(self.status_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_icon.sizePolicy().hasHeightForWidth())
        self.status_icon.setSizePolicy(sizePolicy)
        self.status_icon.setObjectName("status_icon")
        self.horizontalLayout.addWidget(self.status_icon)
        self.status_label = QtWidgets.QLabel(self.status_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_label.sizePolicy().hasHeightForWidth())
        self.status_label.setSizePolicy(sizePolicy)
        self.status_label.setText("")
        self.status_label.setWordWrap(True)
        self.status_label.setObjectName("status_label")
        self.horizontalLayout.addWidget(self.status_label)
        self.verticalLayout.addWidget(self.status_box)
        self.jobs_view = QtWidgets.QTableView(self.centralwidget)
        self.jobs_view.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.jobs_view.setAlternatingRowColors(True)
        self.jobs_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.jobs_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.jobs_view.setShowGrid(False)
        self.jobs_view.setSortingEnabled(True)
        self.jobs_view.setWordWrap(False)
        self.jobs_view.setObjectName("jobs_view")
        self.jobs_view.horizontalHeader().setHighlightSections(False)
        self.jobs_view.horizontalHeader().setSortIndicatorShown(True)
        self.jobs_view.horizontalHeader().setStretchLastSection(True)
        self.jobs_view.verticalHeader().setVisible(False)
        self.verticalLayout.addWidget(self.jobs_view)
        StatusWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(StatusWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 851, 25))
        self.menubar.setObjectName("menubar")
        self.menuCuteBorg_Status = QtWidgets.QMenu(self.menubar)
        self.menuCuteBorg_Status.setObjectName("menuCuteBorg_Status")
        StatusWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(StatusWindow)
        self.statusbar.setObjectName("statusbar")
        StatusWindow.setStatusBar(self.statusbar)
        self.action_close = QtWidgets.QAction(StatusWindow)
        self.action_close.setObjectName("action_close")
        self.action_quit = QtWidgets.QAction(StatusWindow)
        self.action_quit.setObjectName("action_quit")
        self.action_quit_and_stop = QtWidgets.QAction(StatusWindow)
        self.action_quit_and_stop.setObjectName("action_quit_and_stop")
        self.menuCuteBorg_Status.addAction(self.action_close)
        self.menuCuteBorg_Status.addAction(self.action_quit)
        self.menuCuteBorg_Status.addAction(self.action_quit_and_stop)
        self.menubar.addAction(self.menuCuteBorg_Status.menuAction())

        self.retranslateUi(StatusWindow)
        QtCore.QMetaObject.connectSlotsByName(StatusWindow)

    def retranslateUi(self, StatusWindow):
        _translate = QtCore.QCoreApplication.translate
        StatusWindow.setWindowTitle(_translate("StatusWindow", "CuteBorg Status"))
        self.status_icon.setText(_translate("StatusWindow", "TextLabel"))
        self.menuCuteBorg_Status.setTitle(_translate("StatusWindow", "&CuteBorg Status"))
        self.action_close.setText(_translate("StatusWindow", "&Close"))
        self.action_close.setToolTip(_translate("StatusWindow", "Close the status window without stopping CuteBorg Status or CuteBorg itself"))
        self.action_close.setShortcut(_translate("StatusWindow", "Ctrl+W"))
        self.action_quit.setText(_translate("StatusWindow", "&Quit CuteBorg Status…"))
        self.action_quit.setToolTip(_translate("StatusWindow", "Close this window and stop CuteBorg Status, leaving CuteBorg itself running."))
        self.action_quit.setShortcut(_translate("StatusWindow", "Ctrl+Q"))
        self.action_quit_and_stop.setText(_translate("StatusWindow", "Q&uit CuteBorg…"))
        self.action_quit_and_stop.setToolTip(_translate("StatusWindow", "Quit CuteBorg Status and stop CuteBorg"))
        self.action_quit_and_stop.setShortcut(_translate("StatusWindow", "Ctrl+Shift+Q"))




#!/usr/bin/env python3
import os.path
import runpy
import sys

import setuptools
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    'borgbackup',
    'pytz',
]

try:
    # this is *super* ugly, but for some reason pip/pkg_resources doesn’t find
    # PyQt5 installed by apt (but it does find borgbackup. super weird.)
    import PyQt5  # NOQA
except ImportError:
    install_requires.append("PyQt5")

setup(
    name="cuteborg",
    version="0.1.1",
    description="Third-party scheduler and UI frontend for Borgbackup",
    long_description=long_description,
    url="https://github.com/horazont/cuteborg",
    author="Jonas Wielicki",
    author_email="jonas@wielicki.name",
    license="GPLv3+",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: POSIX",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
    keywords="borgbackup",
    install_requires=install_requires,
    packages=find_packages(exclude=["tests*"]),
    entry_points={
        "console_scripts": [
            "cuteborg-scheduler=cuteborg.scheduler.cli:main",
            "cuteborg=cuteborg.cli:main",
            "qtborg-trayicon=qtborg.trayicon.cli:main",
        ]
    }
)



#pylint: skip-file
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'data/ui/manager/ui_new_repo_wizard.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Wizard(object):
    def setupUi(self, Wizard):
        Wizard.setObjectName("Wizard")
        Wizard.resize(559, 504)
        self.page_type = QtWidgets.QWizardPage()
        self.page_type.setObjectName("page_type")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.page_type)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioButton = QtWidgets.QRadioButton(self.page_type)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.radioButton.setFont(font)
        self.radioButton.setChecked(True)
        self.radioButton.setObjectName("radioButton")
        self.verticalLayout.addWidget(self.radioButton)
        self.label_2 = QtWidgets.QLabel(self.page_type)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setWordWrap(True)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.radioButton_2 = QtWidgets.QRadioButton(self.page_type)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.radioButton_2.setFont(font)
        self.radioButton_2.setObjectName("radioButton_2")
        self.verticalLayout.addWidget(self.radioButton_2)
        self.label_3 = QtWidgets.QLabel(self.page_type)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        Wizard.addPage(self.page_type)
        self.page_local_select_directory = QtWidgets.QWizardPage()
        self.page_local_select_directory.setObjectName("page_local_select_directory")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.page_local_select_directory)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label = QtWidgets.QLabel(self.page_local_select_directory)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_3.addWidget(self.label)
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_4 = QtWidgets.QLabel(self.page_local_select_directory)
        self.label_4.setObjectName("label_4")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.directory = QtWidgets.QLineEdit(self.page_local_select_directory)
        self.directory.setObjectName("directory")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.directory)
        self.verticalLayout_3.addLayout(self.formLayout_2)
        self.groupBox = QtWidgets.QGroupBox(self.page_local_select_directory)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setObjectName("groupBox")
        self.formLayout = QtWidgets.QFormLayout(self.groupBox)
        self.formLayout.setObjectName("formLayout")
        self.label_5 = QtWidgets.QLabel(self.groupBox)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.label_6 = QtWidgets.QLabel(self.groupBox)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.label_7 = QtWidgets.QLabel(self.groupBox)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.label_8 = QtWidgets.QLabel(self.groupBox)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.device_info_type = QtWidgets.QLineEdit(self.groupBox)
        self.device_info_type.setReadOnly(True)
        self.device_info_type.setObjectName("device_info_type")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.device_info_type)
        self.device_info_size = QtWidgets.QLineEdit(self.groupBox)
        self.device_info_size.setReadOnly(True)
        self.device_info_size.setObjectName("device_info_size")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.device_info_size)
        self.device_info_partition_size = QtWidgets.QLineEdit(self.groupBox)
        self.device_info_partition_size.setReadOnly(True)
        self.device_info_partition_size.setObjectName("device_info_partition_size")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.device_info_partition_size)
        self.device_info_free_space = QtWidgets.QLineEdit(self.groupBox)
        self.device_info_free_space.setReadOnly(True)
        self.device_info_free_space.setObjectName("device_info_free_space")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.device_info_free_space)
        self.verticalLayout_3.addWidget(self.groupBox)
        self.groupBox_2 = QtWidgets.QGroupBox(self.page_local_select_directory)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_20 = QtWidgets.QLabel(self.groupBox_2)
        self.label_20.setWordWrap(True)
        self.label_20.setObjectName("label_20")
        self.verticalLayout_4.addWidget(self.label_20)
        self.checkBox = QtWidgets.QCheckBox(self.groupBox_2)
        self.checkBox.setObjectName("checkBox")
        self.verticalLayout_4.addWidget(self.checkBox)
        self.verticalLayout_3.addWidget(self.groupBox_2)
        Wizard.addPage(self.page_local_select_directory)
        self.page_remote_configure = QtWidgets.QWizardPage()
        self.page_remote_configure.setObjectName("page_remote_configure")
        self.formLayout_3 = QtWidgets.QFormLayout(self.page_remote_configure)
        self.formLayout_3.setObjectName("formLayout_3")
        self.label_12 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_12.setObjectName("label_12")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_12)
        self.ssh_host = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_host.setObjectName("ssh_host")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.ssh_host)
        self.label_9 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_9.setObjectName("label_9")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_9)
        self.ssh_user = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_user.setClearButtonEnabled(True)
        self.ssh_user.setObjectName("ssh_user")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.ssh_user)
        self.label_11 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_11.setObjectName("label_11")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_11)
        self.ssh_port = QtWidgets.QSpinBox(self.page_remote_configure)
        self.ssh_port.setMinimum(1)
        self.ssh_port.setMaximum(65535)
        self.ssh_port.setProperty("value", 22)
        self.ssh_port.setObjectName("ssh_port")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.ssh_port)
        self.label_13 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_13.setObjectName("label_13")
        self.formLayout_3.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_13)
        self.ssh_identity_file = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_identity_file.setClearButtonEnabled(True)
        self.ssh_identity_file.setObjectName("ssh_identity_file")
        self.formLayout_3.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.ssh_identity_file)
        self.label_10 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_10.setObjectName("label_10")
        self.formLayout_3.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_10)
        self.ssh_remote_path = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_remote_path.setObjectName("ssh_remote_path")
        self.formLayout_3.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.ssh_remote_path)
        self.label_14 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_14.setObjectName("label_14")
        self.formLayout_3.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_14)
        self.ssh_remote_port = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_remote_port.setClearButtonEnabled(True)
        self.ssh_remote_port.setObjectName("ssh_remote_port")
        self.formLayout_3.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.ssh_remote_port)
        self.label_15 = QtWidgets.QLabel(self.page_remote_configure)
        self.label_15.setObjectName("label_15")
        self.formLayout_3.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_15)
        self.ssh_command_line = QtWidgets.QLineEdit(self.page_remote_configure)
        self.ssh_command_line.setReadOnly(True)
        self.ssh_command_line.setObjectName("ssh_command_line")
        self.formLayout_3.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.ssh_command_line)
        self.test_connection = QtWidgets.QPushButton(self.page_remote_configure)
        self.test_connection.setObjectName("test_connection")
        self.formLayout_3.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.test_connection)
        Wizard.addPage(self.page_remote_configure)
        self.page_pruning = QtWidgets.QWizardPage()
        self.page_pruning.setObjectName("page_pruning")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.page_pruning)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_21 = QtWidgets.QLabel(self.page_pruning)
        self.label_21.setWordWrap(True)
        self.label_21.setObjectName("label_21")
        self.verticalLayout_5.addWidget(self.label_21)
        self.checkBox_3 = QtWidgets.QCheckBox(self.page_pruning)
        self.checkBox_3.setObjectName("checkBox_3")
        self.verticalLayout_5.addWidget(self.checkBox_3)
        self.widget = PruneWidget(self.page_pruning)
        self.widget.setObjectName("widget")
        self.verticalLayout_5.addWidget(self.widget)
        self.widget.raise_()
        self.checkBox_3.raise_()
        self.label_2.raise_()
        Wizard.addPage(self.page_pruning)
        self.page_encryption = QtWidgets.QWizardPage()
        self.page_encryption.setObjectName("page_encryption")
        self.formLayout_4 = QtWidgets.QFormLayout(self.page_encryption)
        self.formLayout_4.setObjectName("formLayout_4")
        self.label_17 = QtWidgets.QLabel(self.page_encryption)
        self.label_17.setObjectName("label_17")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_17)
        self.encryption_mode = QtWidgets.QComboBox(self.page_encryption)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.encryption_mode.sizePolicy().hasHeightForWidth())
        self.encryption_mode.setSizePolicy(sizePolicy)
        self.encryption_mode.setObjectName("encryption_mode")
        self.encryption_mode.addItem("")
        self.encryption_mode.addItem("")
        self.encryption_mode.addItem("")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.encryption_mode)
        self.label_18 = QtWidgets.QLabel(self.page_encryption)
        self.label_18.setObjectName("label_18")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_18)
        self.encryption_passphrase = QtWidgets.QLineEdit(self.page_encryption)
        self.encryption_passphrase.setEchoMode(QtWidgets.QLineEdit.Password)
        self.encryption_passphrase.setClearButtonEnabled(False)
        self.encryption_passphrase.setObjectName("encryption_passphrase")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.encryption_passphrase)
        self.encryption_save_passphrase = QtWidgets.QCheckBox(self.page_encryption)
        self.encryption_save_passphrase.setObjectName("encryption_save_passphrase")
        self.formLayout_4.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.encryption_save_passphrase)
        self.label_19 = QtWidgets.QLabel(self.page_encryption)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_19.sizePolicy().hasHeightForWidth())
        self.label_19.setSizePolicy(sizePolicy)
        self.label_19.setWordWrap(True)
        self.label_19.setObjectName("label_19")
        self.formLayout_4.setWidget(3, QtWidgets.QFormLayout.SpanningRole, self.label_19)
        Wizard.addPage(self.page_encryption)
        self.page_finish = QtWidgets.QWizardPage()
        self.page_finish.setObjectName("page_finish")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.page_finish)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_16 = QtWidgets.QLabel(self.page_finish)
        self.label_16.setWordWrap(True)
        self.label_16.setObjectName("label_16")
        self.verticalLayout_2.addWidget(self.label_16)
        self.plainTextEdit = QtWidgets.QPlainTextEdit(self.page_finish)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.plainTextEdit.setFont(font)
        self.plainTextEdit.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.plainTextEdit.setReadOnly(True)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout_2.addWidget(self.plainTextEdit)
        Wizard.addPage(self.page_finish)

        self.retranslateUi(Wizard)
        self.encryption_mode.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(Wizard)

    def retranslateUi(self, Wizard):
        _translate = QtCore.QCoreApplication.translate
        Wizard.setWindowTitle(_translate("Wizard", "Wizard"))
        self.page_type.setTitle(_translate("Wizard", "New repository"))
        self.page_type.setSubTitle(_translate("Wizard", "Select the type of the repository you want to create. If in doubt, a local directory will be the choice you want."))
        self.radioButton.setText(_translate("Wizard", "&Local directory or device"))
        self.label_2.setText(_translate("Wizard", "<html><head/><body><p>A folder which is on a local device, no network involved.</p></body></html>"))
        self.radioButton_2.setText(_translate("Wizard", "Remote server"))
        self.label_3.setText(_translate("Wizard", "<html><head/><body><p>A folder on a server accessible via SSH (SFTP is <span style=\" font-style:italic;\">not</span> sufficient).</p></body></html>"))
        self.page_local_select_directory.setTitle(_translate("Wizard", "Local device"))
        self.page_local_select_directory.setSubTitle(_translate("Wizard", "Select the device and the directory on the device to use."))
        self.label.setText(_translate("Wizard", "<html><head/><body><p>If the device you want to store backups on is removable, connect it now.</p></body></html>"))
        self.label_4.setText(_translate("Wizard", "Directory"))
        self.groupBox.setTitle(_translate("Wizard", "Device information"))
        self.label_5.setText(_translate("Wizard", "Type"))
        self.label_6.setText(_translate("Wizard", "Size"))
        self.label_7.setText(_translate("Wizard", "Partition size"))
        self.label_8.setText(_translate("Wizard", "Free space"))
        self.groupBox_2.setTitle(_translate("Wizard", "Removable device settings"))
        self.label_20.setText(_translate("Wizard", "<html><head/><body><p><span style=\" font-weight:600;\">Note:</span> If the device is not connected when the backup is supposed to run, a notification will inform you that you should plug the device in. Once the device is plugged in, the jobs affecting the device will be scheduled.</p></body></html>"))
        self.checkBox.setToolTip(_translate("Wizard", "<html><head/><body><p>When checked, jobs which are scheduled for the interval during which the device is plugged in are run immediately (with short grace period) instead of at their scheduled time. This is handy for daily backup jobs: simply plug in the device and <span style=\" font-style:italic;\">cuteborg</span> will do its duty.</p></body></html>"))
        self.checkBox.setText(_translate("Wizard", "Allow to run jobs early when the device is being plugged in"))
        self.page_remote_configure.setTitle(_translate("Wizard", "Remote repository"))
        self.page_remote_configure.setSubTitle(_translate("Wizard", "Configure the connection to the remote repository."))
        self.label_12.setText(_translate("Wizard", "Host name"))
        self.label_9.setText(_translate("Wizard", "User"))
        self.label_11.setText(_translate("Wizard", "Port"))
        self.label_13.setText(_translate("Wizard", "Identity file"))
        self.label_10.setText(_translate("Wizard", "Path"))
        self.label_14.setText(_translate("Wizard", "Remote borg executable"))
        self.label_15.setText(_translate("Wizard", "SSH command line"))
        self.test_connection.setText(_translate("Wizard", "Test connection"))
        self.page_pruning.setTitle(_translate("Wizard", "Pruning"))
        self.page_pruning.setSubTitle(_translate("Wizard", "Borg can automatically manage the amount of archives kept. Use the following settings to control the behaviour."))
        self.label_21.setText(_translate("Wizard", "<html><head/><body><p><span style=\" font-weight:600;\">Note:</span> If you choose to use the global settings with the box below, changes to the global settings made <span style=\" font-style:italic;\">after</span> the repository is created will be applied to the repository, too. If you simply want to copy the global settings without future changes to the global settings affecting the repository, uncheck the box below.</p></body></html>"))
        self.checkBox_3.setText(_translate("Wizard", "Use global settings"))
        self.page_encryption.setTitle(_translate("Wizard", "Encryption"))
        self.page_encryption.setSubTitle(_translate("Wizard", "Enable encryption for the repository."))
        self.label_17.setText(_translate("Wizard", "Encryption mode"))
        self.encryption_mode.setItemText(0, _translate("Wizard", "No encryption"))
        self.encryption_mode.setItemText(1, _translate("Wizard", "Save protected key in repository"))
        self.encryption_mode.setItemText(2, _translate("Wizard", "Save protected key only locally"))
        self.label_18.setText(_translate("Wizard", "Passphrase"))
        self.encryption_save_passphrase.setText(_translate("Wizard", "Save passphrase in local keyring"))
        self.label_19.setText(_translate("Wizard", "<html><head/><body><h3>Explanation of the modes</h3><p><span style=\" font-weight:600;\">No encryption:</span> All data is stored without encryption.</p><p>The following options enable client-side encryption: The server never sees file names or file contents in plain-text.</p><p><span style=\" font-weight:600;\">Save protected key in repository:</span> The key is stored inside the repository (in its “config” file). Use this mode if you trust in your good passphrase giving you enough protection. The repository server never sees the plaintext key. This has the advantage that if your local system is unusable, you can still recover using remote repositories with this option.</p><p><span style=\" font-weight:600;\">Save protected key only locally: </span>The key is stored on your local disk (in <span style=\" font-family:\'DejaVu Sans Mono\';\">~/.config/borg/keys/</span>). Use this mode if you want “passphrase and having-the-key” security. <span style=\" font-style:italic;\">Note</span>: If you loose the key stored locally, you cannot recover data from the repository.</p></body></html>"))
        self.page_finish.setTitle(_translate("Wizard", "Finished!"))
        self.page_finish.setSubTitle(_translate("Wizard", "The configuration is finished."))
        self.label_16.setText(_translate("Wizard", "If needed, you can make modifications to the repository configuration below. In case of doubt, simply click continue."))

from qtborg.prune_widget import PruneWidget
